"""Module for agreement and consensus analysis between raters"""

from numpy import (arange, argmax, asarray, concatenate, diff, invert, 
                   logical_and, maximum, mean, minimum, newaxis, ones, repeat, 
                   sum, vstack, where, zeros)

from .. import Graphoelement

class MatchedEvents:
    """Class for storing matched events and producing statistics.
    
    Parameters
    ----------
    tp : ndarray
        true positives as boolean array of shape len(detection) x len(standard)
    fp : ndarray
        indices of false positives in detection
    fn : ndarray
        indices of false negatives in standard
    detection : list of dict
        list of detected events tested against the standard, with 'start', 
        'end' and 'chan'
    standard : list of dict
        list of ground-truth events, with 'start', 'end' and 'chan'
    threshold : float
        minimum intersection-union score for events to be considered 
        overlapping
    """
    def __init__(self, tp, fp, fn, detection, standard, threshold):
        self.tp = tp
        self.fp = fp
        self.fn = fn
        self.detection = detection
        self.standard = standard
        self.threshold = threshold        
        self.n_tp = sum(tp)
        self.n_fp = len(fp)
        self.n_fn = len(fn)

    @property
    def recall(self):
        tp = self.n_tp
        fn = self.n_fn
        if tp + fn == 0:
            return 0
        return tp / (tp + fn)
    
    @property
    def precision(self):
        tp = self.n_tp
        fp = self.n_fp
        if tp + fp == 0:
            return 0
        return tp / (tp + fp)
    
    @property
    def f1score(self):     
        recall = self.recall
        precision = self.precision
        if precision + recall == 0:
            return 0
        return 2 * precision * recall / (precision + recall)
    
    def to_annot(self, annot, category, name, s_freq=512):
        """Write matched events to Wonambi XML file for visualization.
        
        Parameters
        ----------
        annot : instance of Annotations
            Annotations file
        category : str
            'tp_cons', 'tp_det', 'tp_std', 'fp' or 'fn'
        name : str
            name for the event type
        s_freq : int
            sampling frequency, in Hz, only required for 'tp_cons' category
        """
        if 'tp_cons' == category:
            cons = consensus((self.detection, self.standard), 1, s_freq)
            events = cons.events
        
        elif 'tp_det' == category:
            events = asarray(self.detection)[self.tp.any(axis=1)]
            
        elif 'tp_std' == category:
            events = asarray(self.standard)[self.tp.any(axis=0)] 
            
        elif 'fp' == category:
            events = asarray(self.detection)[self.fp]
        
        elif 'fn' == category:
            events = asarray(self.standard)[self.fn]
        
        else:
            raise ValueError("Invalid category.")
        
        annot.add_events(events, name=name, chan=None)
            
    def all_to_annot(self, annot, names=['TPd', 'TPs', 'FP', 'FN']):
        """Convenience function to write all events to XML by category, showing
        overlapping TP detection and TP standard."""
        self.to_annot(annot, 'tp_det', names[0])
        self.to_annot(annot, 'tp_std', names[1])
        self.to_annot(annot, 'fp', names[2])
        self.to_annot(annot, 'fn', names[3])


def consensus(events, threshold, s_freq, min_duration=None, weights=None):
    """Take two or more event lists and output a merged list based on 
    consensus.
    
    Parameters
    ----------
    events: tuple of lists of dict
        two or more lists of events from different raters, with 'start', 'end'
        and 'chan'
    threshold : float
        value between 0 and 1 to threshold consensus. Consensus is computed on
        a per-sample basis. For a given rater, if an event is present at a 
        sample, that rater-sample is assigned the value 1; otherwise it is 
        assigned 0. The arithmetic mean is taken per sample across all raters, 
        and if this mean exceeds 'threshold', the sample is counted as 
        belonging to a merged event.
    s_freq : int
        sampling frequency, in Hz
    min_duration : float, optional
        minimum duration for merged events, in s.
    weights : list of float
        a vector of relative weights of each event type
        
    Returns
    -------
    instance of wonambi.Graphoelement
        events merged by consensus
    """
    chan = [one_rater[0]['chan'] for one_rater in events if one_rater][0]
    beg = min([one_rater[0]['start'] for one_rater in events if one_rater])
    end = max([one_rater[-1]['end'] for one_rater in events if one_rater])
    n_samples = int((end - beg) * s_freq)
    times = arange(beg, end + 1/s_freq, 1/s_freq)
    if weights is None:
        weights = ones(len(events))
    
    positives = zeros((len(events), n_samples))
    for i, (one_rater, wt) in enumerate(zip(events, weights)):
        for ev in one_rater:
            n_start = int((ev['start'] - beg) * s_freq)
            n_end = int((ev['end'] - beg) * s_freq)
            positives[i, n_start:n_end].fill(wt)
                
    consensus = mean(positives, axis=0)
    consensus[consensus >= threshold] = 1
    consensus[consensus < 1] = 0
    consensus = concatenate(([0], consensus, [0]))
    on_off = diff(consensus)
    onsets = where(on_off == 1)
    offsets = where(on_off == -1)
    start_times = times[onsets]
    end_times = times[offsets]
    merged = vstack((start_times, end_times))
    
    if min_duration:
        merged = merged[:, merged[1, :] - merged[0, :] >= min_duration]
        
    out = Graphoelement()
    out.events = [{'start': merged[0, i], 
                   'end': merged[1, i],
                   'chan': chan} for i in range(merged.shape[1])]

    return out     

def consensus_exact(events, threshold, s_freq, window=None, min_duration=None, weights=None):
    """Take two or more event lists and output a merged list based on 
    consensus, where agreement is exactly equal to a threshold.
    This is useful when combining >2 event types, and creating a 
    consensus event type based on some combination of these events.
    
    Parameters
    ----------
    events: tuple of lists of dict
        two or more lists of events from different raters, with 'start', 'end'
        and 'chan'
    threshold : float
        value between 0 and 1 to threshold consensus. Consensus is computed on
        a per-sample basis. For a given rater, if an event is present at a 
        sample, that rater-sample is assigned the value 1; otherwise it is 
        assigned 0. The arithmetic sum is taken per sample across all raters, 
        and if this exactly equals 'threshold', the sample is counted as 
        belonging to a merged event.
    s_freq : int
        sampling frequency, in Hz
    min_duration : float, optional
        minimum duration for merged events, in s.
    weights : a dict containing event names (str) and their corresponding 
        weighting (int) e.g. {'low' : 1,'med' : 2,'high' : 3}
        
    Returns
    -------
    instance of wonambi.Graphoelement
        events merged by consensus and named by confidence rating
        
    Notes
    -----
    This function is a modification of agreement.consensus contributed by 
    Nathan Cross.
    """
    chan = [one_rater[0]['chan'] for one_rater in events if one_rater][0]
    if window is None:
        beg = min([one_rater[0]['start'] for one_rater in events if one_rater])
        end = max([one_rater[-1]['end'] for one_rater in events if one_rater])
    else:
        beg = window[0]
        end = window[1]
   
    n_samples = int((end - beg) * s_freq)
    times = arange(beg, end + 1/s_freq, 1/s_freq)
    if weights is None:
        weights = {'low':1,'med':2,'high':3}
    
    positives = zeros((len(events), n_samples))
    for i, one_rater in enumerate(events):
        for ev in one_rater:
            n_start = int((ev['start'] - beg) * s_freq)
            n_end = int((ev['end'] - beg) * s_freq)
            if ev['name'] == 'low':
                positives[i, n_start:n_end].fill(weights['low'])
            elif ev['name'] == 'med':
                positives[i, n_start:n_end].fill(weights['med'])
            elif ev['name'] == 'high':
                positives[i, n_start:n_end].fill(weights['high'])
                
    consensus = sum(positives, axis=0)
    consensus[consensus != threshold] = 0
    consensus[consensus == threshold] = 1  
    consensus = concatenate(([0], consensus, [0]))
    on_off = diff(consensus)
    onsets = where(on_off == 1)
    offsets = where(on_off == -1)
    start_times = times[onsets]
    end_times = times[offsets]
    merged = vstack((start_times, end_times))
    
    if min_duration:
        merged = merged[:, merged[1, :] - merged[0, :] >= min_duration]
        
    out = Graphoelement()
    out.events = [{'start': merged[0, i], 
                   'end': merged[1, i],
                   'chan': chan} for i in range(merged.shape[1])]

    return out 
            
def match_events(detection, standard, threshold):
    """Find best matches between detected and standard events, by a thresholded
    intersection-union rule.
    
    Parameters
    ----------
    detection : list of dict
        list of detected events to be tested against the standard, with 
        'start', 'end' and 'chan'
    standard : list of dict
        list of ground-truth events, with 'start', 'end' and 'chan'
    threshold : float
        minimum intersection-union score to match a pair, between 0 and 1
        
    Returns
    -------
    instance of MatchedEvents
        indices of true positives, false positives and false negatives, with
        statistics (recall, precision, F1)
    """
    # Vectorize start and end times and set up for broadcasting
    det_beg = asarray([x['start'] for x in detection])[:, newaxis]
    det_end = asarray([x['end'] for x in detection])[:, newaxis]
    std_beg = asarray([x['start'] for x in standard])[newaxis, :]
    std_end = asarray([x['end'] for x in standard])[newaxis, :]

    # Get durations and broadcast them
    det_dur = repeat(det_end - det_beg, len(standard), axis=1)
    std_dur = repeat(std_end - std_beg, len(detection), axis=0)
    
    # Subtract every end by every start and find overlaps
    det_minus_std = det_end - std_beg # array of shape (len(det), len(std))
    std_minus_det = std_end - det_beg    
    overlapping = logical_and(det_minus_std > 0, std_minus_det > 0)
    
    # Find intersection and union
    shorter_diff = minimum(det_minus_std, std_minus_det)
    longer_diff = maximum(det_minus_std, std_minus_det)
    
    shorter_dur = minimum(det_dur, std_dur)
    longer_dur = maximum(det_dur, std_dur)
    
    interx = minimum(shorter_diff, shorter_dur)
    union = maximum(longer_diff, longer_dur)
        
    # Compute intersection-union score and set non-overlapping pairs to 0
    iu = interx / union
    iu[invert(overlapping)] = 0
    
    # Threshold IU score to yield  True Positive candidates
    iu[iu <= threshold] = 0
    
    # If no events, tp and fp are empty, fn is all events
    if iu.size == 0:
        tp = fp = asarray([])
        fn = arange(len(standard))
    else:
    
        # Find partial matches, round 1
        det_match1 = argmax(iu, axis=1)
        std_match1 = argmax(iu, axis=0)
        
        # Find full matches, round 1, then remove them from IU
        tp = zeros(iu.shape, dtype=bool)
        for i, j in enumerate(std_match1):
            if det_match1[j] == i:
                tp[j, i] = True
                iu[j, :].fill(0)
                iu[:, i].fill(0)
        
        # Round 2
        det_match2 = argmax(iu, axis=1)
        std_match2 = argmax(iu, axis=0)
        
        for i, j in enumerate(std_match2):
            if det_match2[j] == i:
                tp[j, i] = True
    
        # Find false positives and false negatives
        fp = where(logical_and(det_match1 == 0, det_match2 == 0))[0]
        fn = where(logical_and(std_match1 == 0, std_match2 == 0))[0]
    
    # Store in MatchedEvents class, which computes statistics
    match = MatchedEvents(tp, fp, fn, detection, standard, threshold)
    
    return match



"""Module contains the different types of data formats.

The main class is Data, all the other classes should depend on it. The other
classes are given only as convenience, but they should not overwride
Data.__call__, which needs to be very general.
"""
from collections import OrderedDict
from collections.abc import Iterable
from copy import deepcopy
from logging import getLogger
from pathlib import Path

from numpy import arange, array, empty, ix_, nan, squeeze, where

lg = getLogger()


class Data:
    """General class containing recordings.

    Parameters
    ----------
    data : ndarray
        one matrix with dimension matching the number of axes. You can pass
        only one trial.
    s_freq : int
        sampling frequency
    axes : dict
        dictionary where the key is the name of the axis and the values must be
        a numpy vector with the actual values.

    Attributes
    ----------
    data : ndarray (dtype='O')
        the data as trials. Each trial is a ndarray (dtype='d' or 'f')
    axis : OrderedDict
        dictionary with axiss (standard names are 'chan', 'time', 'freq');
        values should be numpy array
    s_freq : int
        sampling frequency
    start_time : instance of datetime.datetime
        the start time of the recording
    attr : dict
        contains additional information about the dataset, with keys:
            - surf
            - chan
            - scores

    Notes
    -----
    Something which is not immediately clear for chan. dtype='U' (meaning
    Unicode) actually creates string of type str\_, while if you use dtype='S'
    (meaning String) it creates strings of type bytes\_.
    """
    def __init__(self, data=None, s_freq=None, **kwargs):

        self.s_freq = s_freq

        if data is None:
            self.data = array([], dtype='O')
        else:
            self.data = array((1, ), dtype='O')
            self.data[0] = data

        self.axis = OrderedDict()
        if data is not None:
            """temporary solution until PEP0468
            kwargs is a dict, so no order. We try to reconstruct order based
            on number of values for each value, but not 100% reliable.
            """
            count_kwargs = {len(v): k for k, v in kwargs.items()}
            if len(set(count_kwargs)) != len(list(count_kwargs)):
                lg.warning('Some arguments have the same length, so the order '
                           'of the axes might be incorrect')
            axes = OrderedDict()
            for n_dim in data.shape:
                try:
                    axis_with_right_ndim = count_kwargs[n_dim]
                    axes[axis_with_right_ndim] = kwargs[axis_with_right_ndim]
                except KeyError:
                    raise ValueError('Number of dimensions in axis does not '
                                     'match number of dimensions in data')

            for axis, value in axes.items():
                self.axis[axis] = array((1,), dtype='O')
                self.axis[axis][0] = value

        self.start_time = None

        self.attr = {'surf': None,
                     'chan': None,
                     'scores': None,
                     }

    def __call__(self, trial=None, tolerance=None, **axes):
        """Return the recordings and their time stamps.

        Parameters
        ----------
        trial : list of int or ndarray (dtype='i') or int
            which trials you want (if it's one int, it returns the actual
            matrix).
        **axes
            Arbitrary axiss to select from. You specify the axis and
            the values as list or tuple of the values that you want.
        tolerance : float
            if one of the axiss is a number, it specifies the tolerance to
            consider one value as chosen (take into account floating-precision
            errors).

        Returns
        -------
        ndarray
            ndarray containing the data with the same number of axiss as
            the original data. The length of the axis is equal to the
            length of the data, UNLESS you specify an axis with values. In
            that case, the length is equal to the values that you want.

            If you specify only one trial (as int, not as tuple or list), then
            it returns the actual matrix. Otherwise, it returns a ndarray
            (dtype='O') of length equal to the trials.

        Notes
        -----
        You cannot specify intervals here, you can do it in Select.

        """
        if trial is None:
            trial = range(self.number_of('trial'))

        squeeze_trial = False
        try:
            iter(trial)
        except TypeError:  # 'int' object is not iterable
            trial = (trial, )
            squeeze_trial = True

        output = empty(len(trial), dtype='O')

        for cnt, i in enumerate(trial):

            output_shape = []
            idx_data = []
            idx_output = []
            squeeze_axis = []

            for axis, values in self.axis.items():
                if axis in axes.keys():
                    selected_values = axes[axis]
                    if (isinstance(selected_values, Iterable) and
                        not isinstance(selected_values, str)):
                        n_values = len(selected_values)
                    else:
                        n_values = 1
                        selected_values = array([selected_values])
                        squeeze_axis.append(self.index_of(axis))

                    idx = _get_indices(values[i],
                                       selected_values,
                                       tolerance=tolerance)
                    if len(idx[0]) == 0:
                        lg.warning('No index was selected for ' + axis)

                    idx_data.append(idx[0])
                    idx_output.append(idx[1])
                else:
                    n_values = len(values[i])
                    idx_data.append(arange(n_values))
                    idx_output.append(arange(n_values))

                output_shape.append(n_values)

            output[cnt] = empty(output_shape, dtype=self.data[i].dtype)
            output[cnt].fill(nan)

            if all([len(x) > 0 for x in idx_data]):
                ix_output = ix_(*idx_output)
                ix_data = ix_(*idx_data)
                output[cnt][ix_output] = self.data[i][ix_data]

            if len(squeeze_axis) > 0:
                output[cnt] = squeeze(output[cnt],
                                      axis=tuple(squeeze_axis))

        if squeeze_trial:
            output = output[0]

        return output

    @property
    def list_of_axes(self):
        """Return the name of all the axes in the data."""
        return tuple(self.axis.keys())

    def index_of(self, axis):
        """Return the index of a axis.

        Parameters
        ----------
        axis : str
            Name of the axis (such as 'trial', 'time', etc)

        Returns
        -------
        int or ndarray (dtype='int')
            number of trial (as int) or number of element in the selected
            axis (if any of the other axiss) as 1d array.

        Raises
        ------
        ValueError
            If the requested axis is not in the data.

        """
        return list(self.axis.keys()).index(axis)

    def number_of(self, axis):
        """Return the number of in one axis, as generally as possible.

        Parameters
        ----------
        axis : str
            Name of the axis (such as 'trial', 'time', etc)

        Returns
        -------
        int or ndarray (dtype='int')
            number of trial (as int) or number of element in the selected
            axis (if any of the other axiss) as 1d array.

        Raises
        ------
        KeyError
            If the requested axis is not in the data.

        Notes
        -----
        or is it better to catch the exception?

        """
        if axis == 'trial':
            return len(self.data)
        else:
            n_trial = self.number_of('trial')
            output = empty(n_trial, dtype='int')
            for i in range(n_trial):
                output[i] = len(self.axis[axis][i])

            return output

    def __getattr__(self, possible_axis):
        """Return the axis with a shorter syntax.

        Parameters
        ----------
        possible_axis : str
            one of the axes

        Returns
        -------
        value of the axis of interest

        Notes
        ------
        The if-statement "startswith" is necessary to avoid recursionerror
        when loading the class.
        """
        if possible_axis.startswith('__'):
            raise AttributeError(possible_axis)

        try:
            return self.axis[possible_axis]
        except KeyError:
            raise AttributeError(possible_axis)

    def __iter__(self):
        """Implement generator for each trial.

        The generator returns the data for each trial. This is of course really
        convenient for map and parallel processing.

        Examples
        --------
        >>> from wonambi.trans import math
        >>> for one_trial in iter(data):
        >>>     one_mean = math(one_trial, operator_name='mean', axis='time')
        >>>     print(one_mean.data[0])
        """
        for trial in range(self.number_of('trial')):

            output = self._copy(axis=False)

            for one_axis in self.axis:
                output.axis[one_axis] = empty(1, dtype='O')
            output.data = empty(1, dtype='O')

            output.data[0] = self.data[trial]
            for one_axis in output.axis:
                output.axis[one_axis][0] = self.axis[one_axis][trial]

            yield output

    def _copy(self, axis=True, attr=True, data=False):
        """Create a new instance of Data, but does not copy the data
        necessarily.

        Parameters
        ----------
        axis : bool, optional
            deep copy the axes (default: True)
        attr : bool, optional
            deep copy the attributes (default: True)
        data : bool, optional
            deep copy the data (default: False)

        Returns
        -------
        instance of Data (or ChanTime, ChanFreq, ChanTimeFreq)
            copy of the data, but without the actual data

        Notes
        -----
        It's important that we copy all the relevant information here. If you
        add new attributes, you should add them here.

        Remember that it deep-copies all the information, so if you copy data
        the size might become really large.
        """
        cdata = type(self)()  # create instance of the same class

        cdata.s_freq = self.s_freq
        cdata.start_time = self.start_time

        if axis:
            cdata.axis = deepcopy(self.axis)
        else:
            cdata_axis = OrderedDict()
            for axis_name in self.axis:
                cdata_axis[axis_name] = array([], dtype='O')
            cdata.axis = cdata_axis

        if attr:
            cdata.attr = deepcopy(self.attr)

        if data:
            cdata.data = deepcopy(self.data)

        else:
            # empty data with the correct number of trials
            cdata.data = empty(self.number_of('trial'), dtype='O')

        return cdata

    def export(self, filename, export_format='FieldTrip', **options):
        """Export data in other formats.

        Parameters
        ----------
        filename : path to file
            file to write
        export_format : str, optional
            supported export format is currently FieldTrip, EDF, FIFF, Wonambi,
            BrainVision

        Notes
        -----
        'edf' takes an optional argument "physical_max", see write_edf.

        'wonambi' takes an optional argument "subj_id", see write_wonambi.
        wonambi format creates two files, one .won with the dataset info as json
        file and one .dat with the memmap recordings.

        'brainvision' takes an additional argument ("markers") which is a list
        of dictionaries with fields:
            "name" : str (name of the marker),
            "start" : float (start time in seconds)
            "end" : float (end time in seconds)

        'bids' has an optional argument "markers", like in 'brainvision'
        """
        filename = Path(filename)
        filename.parent.mkdir(parents=True, exist_ok=True)

        export_format = export_format.lower()
        if export_format == 'edf':
            from .ioeeg import write_edf  # avoid circular import
            write_edf(self, filename, **options)

        elif export_format == 'fieldtrip':
            from .ioeeg import write_fieldtrip  # avoid circular import
            write_fieldtrip(self, filename)

        elif export_format == 'mnefiff':

            from .ioeeg import write_mnefiff
            write_mnefiff(self, filename)

        elif export_format == 'wonambi':
            from .ioeeg import write_wonambi
            write_wonambi(self, filename, **options)

        elif export_format == 'brainvision':
            from .ioeeg import write_brainvision
            write_brainvision(self, filename, **options)

        elif export_format == 'bids':
            from .ioeeg import write_bids
            write_bids(self, filename, **options)

        else:
            raise ValueError('Cannot export to ' + export_format)


class ChanTime(Data):
    """Specific class for chan-time recordings, with axes:

    chan : ndarray (dtype='O')
        for each trial, channels in the data (dtype='U')
    time : ndarray (dtype='O')
        for each trial, 1d matrix with the time stamp (dtype='f')

    """
    def __init__(self):
        super().__init__()
        self.axis['chan'] = array([], dtype='O')
        self.axis['time'] = array([], dtype='O')


class ChanFreq(Data):
    """Specific class for channel-frequency recordings, with axes:

    chan : ndarray (dtype='O')
        for each trial, channels in the data (dtype='U')
    freq : ndarray (dtype='O')
        for each trial, 1d matrix with the frequency (dtype='f')

    Notes
    -----
    Conceptually, it is reasonable that each trial has the same frequency band,
    so it might be more convenient to use only one array, but it can happen
    that different trials have different frequency bands, so we keep the format
    more open.

    """
    def __init__(self):
        super().__init__()
        self.axis['chan'] = array([], dtype='O')
        self.axis['freq'] = array([], dtype='O')


class ChanTimeFreq(Data):
    """Specific class for channel-time-frequency representation, with axes:

    chan : ndarray (dtype='O')
        for each trial, channels in the data (dtype='U')
    time : ndarray (dtype='O')
        for each trial, 1d matrix with the time stamp (dtype='f')
    freq : ndarray (dtype='O')
        for each trial, 1d matrix with the frequency (dtype='f')

    """
    def __init__(self):
        super().__init__()
        self.axis['chan'] = array([], dtype='O')
        self.axis['time'] = array([], dtype='O')
        self.axis['freq'] = array([], dtype='O')


def _get_indices(values, selected, tolerance):
    """Get indices based on user-selected values.

    Parameters
    ----------
    values : ndarray (any dtype)
        values present in the axis.
    selected : ndarray (any dtype) or tuple or list
        values selected by the user
    tolerance : float
        avoid rounding errors.

    Returns
    -------
    idx_data : list of int
        indices of row/column to select the data
    idx_output : list of int
        indices of row/column to copy into output

    Notes
    -----
    This function is probably not very fast, but it's pretty robust. It keeps
    the order, which is extremely important.

    If you use values in the self.axis, you don't need to specify tolerance.
    However, if you specify arbitrary points, floating point errors might
    affect the actual values. Of course, using tolerance is much slower.

    Maybe tolerance should be part of Select instead of here.

    """
    idx_data = []
    idx_output = []
    for idx_of_selected, one_selected in enumerate(selected):

        if tolerance is None or values.dtype.kind == 'U':
            idx_of_data = where(values == one_selected)[0]
        else:
            idx_of_data = where(abs(values - one_selected) <= tolerance)[0] # actual use min

        if len(idx_of_data) > 0:
            idx_data.append(idx_of_data[0])
            idx_output.append(idx_of_selected)

    return idx_data, idx_output



from numpy import arange
from numpy.random import seed
from numpy.testing import assert_array_equal, assert_array_almost_equal
from pytest import approx, raises

from wonambi import Dataset
from wonambi.attr import Annotations
from wonambi.utils import create_data
from wonambi.trans import select, resample, frequency, get_times, fetch
from wonambi.trans.select import _create_subepochs

from .paths import (annot_psg_path,
                    gui_file,
                    )

seed(0)
data = create_data(n_trial=5)


def test_select_typeerror():
    with raises(TypeError):
        select(data, trial=1)

    with raises(TypeError):
        select(chan='chan01')

    with raises(TypeError):
        select(data, time=1, chan=('chan01', ))


def test_select_trial():

    data1 = select(data, trial=(1, 2))
    assert data1.number_of('trial') == 2

    data1 = select(data, trial=(0, 0))
    assert_array_equal(data1.data[0], data1.data[1])
    assert len(data1.axis['chan']) == 2
    assert data1.number_of('trial') == 2


def test_select_string_selection():
    data1 = select(data, chan=['chan02'])
    assert data1.axis['chan'][0][0] == 'chan02'
    assert data1.data[0].shape[0] == 1


def test_select_empty_selection():
    data1 = select(data, chan=[])
    assert len(data1.axis['chan'][0]) == 0
    assert data1.data[0].shape[0] == 0


def test_select_create_subepochs():

    # 1d
    x = arange(1000)
    nperseg = 100
    step = 50
    v = _create_subepochs(x, nperseg, step)
    assert v.shape == (19, nperseg)
    assert v[0, step] == v[1, 0]

    # 2d
    x = arange(1000).reshape(20, 50)
    nperseg = 10
    step = 5
    v = _create_subepochs(x, nperseg, step)
    assert v.shape == (20, 9, nperseg)
    assert_array_equal(v[:, 0, step], v[:, 1, 0])

    # 3d
    x = arange(1000).reshape(4, 5, 50)
    v = _create_subepochs(x, nperseg, step)
    assert v.shape == (4, 5, 9, nperseg)
    assert_array_equal(v[..., 0, step], v[..., 1, 0])


def test_select_trials():
    data1 = select(data, trial=(1, 4))
    assert_array_equal(data.data[1], data1.data[0])


def test_select_trials_and_string():
    data1 = select(data, trial=(1, 4), chan=('chan01', 'chan02'))
    assert len(data1.axis['chan']) == 2
    assert len(data1.axis['chan'][0]) == 2


def test_select_trials_and_string_invert():
    data1 = select(data, trial=(1, 4), chan=('chan01', 'chan02'), invert=True)
    assert len(data1.axis['chan']) == data.number_of('trial') - 2
    assert len(data1.axis['chan'][0]) == data.number_of('chan')[0] - 2


def test_select_interval():
    data1 = select(data, time=(0.2, 0.5))

    assert data1.axis['time'][0].shape[0] == 76
    assert data1.data[0].shape[1] == 76
    assert data1.data[-1].shape[1] == 76


def test_select_interval_invert():
    data1 = select(data, time=(0.2, 0.5), invert=True)
    assert data1.number_of('time')[0] == data.number_of('time')[0] - 76
    assert data1.data[0].shape[1] == data.number_of('time')[0] - 76
    assert data1.data[-1].shape[1] == data.number_of('time')[0] - 76

    data2 = select(data1, time=(0.2, 0.5), invert=True)
    assert data1.number_of('trial') == data2.number_of('trial')


def test_select_interval_not_in_data():
    data1 = select(data, time=(10.2, 10.5))

    assert len(data1.axis['time'][0]) == 0
    assert data1.data[0].shape[1] == 0
    assert data1.data[-1].shape[1] == 0


def test_select_oneside_interval_0():
    data1 = select(data, time=(None, 0.5))

    assert len(data1.axis['time'][0]) * 2 == len(data.axis['time'][0])
    assert data1.data[0].shape[1] * 2 == data.data[0].shape[1]
    assert data1.axis['time'][0][0] == 0
    assert data1.axis['time'][0][-1] < .5


def test_select_oneside_interval_1():
    data1 = select(data, time=(0.5, None))

    assert len(data1.axis['time'][0]) * 2 == len(data.axis['time'][0])
    assert data1.data[0].shape[1] * 2 == data.data[0].shape[1]
    assert data1.axis['time'][0][0] >= 0.5
    assert data1.axis['time'][0][-1] == data.axis['time'][0][-1]


def test_select_oneside_interval_both():
    data1 = select(data, time=(None, None))

    assert len(data1.axis['time'][0]) == len(data.axis['time'][0])
    assert data1.data[0].shape[1] == data.data[0].shape[1]
    assert data1.axis['time'][0][0] == data.axis['time'][0][0]
    assert data1.axis['time'][0][-1] == data.axis['time'][0][-1]


def test_resample():
    seed(0)
    data = create_data(n_trial=1, s_freq=1024, signal='sine', sine_freq=20)

    NEW_FREQ = 256
    data1 = resample(data, s_freq=NEW_FREQ)

    assert data1.s_freq == NEW_FREQ
    assert data1.data[0].shape[1] == data1.number_of('time')[0]

    freq = frequency(data, taper='boxcar')
    freq1 = frequency(data1, taper='boxcar')
    assert_array_almost_equal(sum(freq.data[0][0, :]),
                              sum(freq1.data[0][0, :]),
                              4)
    
def test_get_times():
    annot = Annotations(str(annot_psg_path))
    
    bundles = get_times(annot, stage=['NREM2'], exclude=True)
    assert len(bundles[0]['times']) == 229
    assert bundles[0]['stage'] == 'NREM2'
    
def test_fetch():
    dset = Dataset(str(gui_file))
    annot = Annotations(str(annot_psg_path))
    
    seg = fetch(dset, annot, stage=['NREM2', 'NREM3'], epoch='locked',
                         reject_epoch=True, reject_artf=True)
    assert len(seg) == 356
    
    seg = fetch(dset, annot, cat=(0, 1, 0, 0), 
                         stage=['NREM2', 'NREM3'], reject_epoch=True, 
                         reject_artf=True)
    assert len(seg) == 31
    assert seg[14]['times'][0] == (34380, 34410)
    
    seg = fetch(dset, annot, cat=(0, 0, 1, 0), stage=['NREM1'])
    seg.read_data(['EEG Fpz-Cz'], ref_chan=['EEG Pz-Oz'])
    assert seg[0]['data']()[0][0].shape == (297000,)
    assert approx(seg[0]['data']()[0][0][100]) == -4.3201466  



"""Package to transform the Data.

The modules in this package are classes (and maybe functions) that transform
the data into data again (all the classes defined in datatype.py). Do not
use this package to transform to other classes. If you want to transform to
basic elements, use the package "detect" for example.

"""
from .filter import filter_, convolve
from .select import (select, resample, get_times, _select_channels, fetch,
                     Segments)
from .frequency import frequency, timefrequency, band_power
from .merge import concatenate
from .math import math, get_descriptives
from .montage import montage, create_virtual_channel
from .peaks import peaks
from .reject import rejectbadchan, remove_artf_evts
from .analyze import export_freq, export_freq_band
from .baseline import apply_baseline



"""Module to detect spindles.
"""
from logging import getLogger
from numpy import (absolute, arange, argmax, argmin, around, asarray,
                   concatenate, cos, diff, exp, empty, histogram,
                   hstack, insert, invert, log10, logical_and, mean, median,
                   nan, ones, percentile, pi, ptp, real, sqrt, square, std,
                   sum, vstack, where, zeros)
from numpy.fft import rfftfreq
from scipy.ndimage.filters import gaussian_filter
from scipy.signal import (argrelmax, butter, cheby2, filtfilt,
                          fftconvolve, hilbert, periodogram, remez,
                          sosfiltfilt, spectrogram)
from scipy.signal.windows import tukey
from scipy.fftpack import next_fast_len
try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QProgressDialog
except ImportError:
    pass

from ..graphoelement import Spindles

lg = getLogger(__name__)
MAX_FREQUENCY_OF_INTEREST = 50
MAX_DURATION = 10


class DetectSpindle:
    """Design spindle detection on a single channel.

    Parameters
    ----------
    method : str
        one of the predefined methods
    frequency : tuple of float
        low and high frequency of spindle band
    duration : tuple of float
        min and max duration of spindles
    merge : bool
        if True, then after events are detected on every channel, events on
        different channels that are separated by less than min_interval will be
        merged into a single event, with 'chan' = the chan of the earlier-onset
        event.

    Attributes
    ----------
    tolerance : float
        during detection and prior to applying the duration criterion,
        candidate events separated by less than this time interval are merged.
        In this way, the detector becomes tolerant to short dips below the
        eligibility threshold (e.g. if the spindle power drops for a split
        second).
    min_interval : float
        after the duration criterion is applied, events separated by less than
        this interval are merged into a single event, with 'chan' = the chan
        of the earlier-onset event.
    power_peaks : str or None
        for peak power statistics. 'peak' or 'interval'. If None, values will
        all be NaN

    Notes
    -----
    See individual detect_* functions for other attribute descriptions.
    """
    def __init__(self, method='Moelle2011', frequency=None, duration=None,
                 merge=False):

        self.method = method
        self.frequency = frequency
        self.merge = merge
        self.tolerance = 0
        self.min_interval = 0
        self.power_peaks = 'interval'
        self.rolloff = None

        if method == 'Ferrarelli2007':
            if self.frequency is None:
                self.frequency = (11, 15)
            self.duration = (0.3, 3)
            self.det_remez = {'freq': self.frequency,
                              'rolloff': 0.9,
                              'dur': 2.56
                              }
            self.det_thresh = 8
            self.sel_thresh = 2

        elif method == 'Moelle2011':
            if self.frequency is None:
                self.frequency = (12, 15)
            self.duration = (0.5, 3)
            self.det_remez = {'freq': self.frequency,
                              'rolloff': 1.7,
                              'dur': 2.36
                               }
            self.moving_rms = {'dur': .2,
                               'step': None}
            self.smooth = {'dur': .2,
                           'win': 'flat'}
            self.det_thresh = 1.5

        elif method == 'Nir2011':
            if self.frequency is None:
                self.frequency = (9.2, 16.8)
            self.duration = (0.5, 2)
            self.det_butter = {'order': 2,
                               'freq': self.frequency,
                               }
            self.tolerance = 1
            self.smooth = {'dur': .04}  # is in fact sigma
            self.det_thresh = 3
            self.sel_thresh = 1


        elif method == 'Wamsley2012':
            if self.frequency is None:
                self.frequency = (12, 15)
            self.duration = (0.3, 3)
            self.det_wavelet = {'f0': mean(self.frequency),
                                'sd': .8,
                                'dur': 1.,
                                'output': 'complex'
                                }
            self.smooth = {'dur': .1,
                           'win': 'flat'}
            self.det_thresh = 4.5

        elif method == 'Martin2013':
            if self.frequency is None:
                self.frequency = (11.5, 14.5)
            self.duration = (.5, 3)
            self.det_remez = {'freq': self.frequency,
                              'rolloff': 1.1,
                              'dur': 2.56
                               }
            self.moving_rms = {'dur': .25,
                               'step': .25}
            self.det_thresh = 95

        elif method == 'Ray2015':
            if self.frequency is None:
                self.frequency = (11, 16)
            self.duration = (.49, None)
            self.cdemod = {'freq': mean(self.frequency)}
            self.det_butter = {'freq': (0.3, 35),
                               'order': 4}
            self.det_low_butter = {'freq': 5,
                                   'order': 4}
            self.min_interval = 0.25 # they only start looking again after .25s
            self.smooth = {'dur': 2 / self.cdemod['freq'],
                           'win': 'triangle'}
            self.zscore = {'dur': 60,
                           'step': None,
                           'pcl_range': None}
            self.det_thresh = 2.33
            self.sel_thresh = 0.1

        elif method == 'Lacourse2018':
            if self.frequency is None:
                self.frequency = (11, 16)
            self.duration = (.3, 2.5)
            self.det_butter = {'freq': self.frequency,
                               'order': 20}
            self.det_butter2 = {'freq': (.3, 30),
                                'order': 5}
            self.windowing = win = {'dur': .3,
                                    'step': .1}
            self.moving_ms = {'dur': win['dur'],
                              'step': win['step']}
            self.moving_power_ratio = {'dur': win['dur'],
                                     'step': win['step'],
                                     'freq_narrow': self.frequency,
                                     'freq_broad': (4.5, 30),
                                     'fft_dur': 2}
            self.zscore = {'dur': 30,
                           'step': None,
                           'pcl_range': (10, 90)}
            self.moving_covar = {'dur': win['dur'],
                                 'step': win['step']}
            self.moving_sd = {'dur': win['dur'],
                              'step': win['step']}
            self.smooth = {'dur': 0.3,
                           'win': 'flat_left'}
            self.abs_pow_thresh = 1.25
            self.rel_pow_thresh = 1.6
            self.covar_thresh = 1.3
            self.corr_thresh = 0.69

        elif 'FASST' in method:
            if self.frequency is None:
                self.frequency = (11, 18)
            self.duration = (.4, 1.3)
            self.det_butter = {'freq': self.frequency,
                               'order': 4}
            self.min_interval = 1
            self.moving_rms = {'dur': .1,
                               'step': None}
            self.smooth = {'dur': .1,
                           'win': 'flat'}
            self.det_thresh = 90

        elif method == 'UCSD':
            if self.frequency is None:
                self.frequency = (10, 16)
            self.duration = (0.3, 3)
            self.det_wavelet = {'freqs': arange(self.frequency[0],
                                                self.frequency[1] + .5, .5),
                                'dur': 1,
                                'width': .5,
                                'win': .5,
                                'sd': None
                                }
            self.det_thresh = 2  # wavelet_peak_thresh
            self.sel_wavelet = {'freqs': arange(self.frequency[0],
                                                self.frequency[1] + .5, .5),
                                'dur': 1,
                                'width': .2,
                                'win': .2,
                                }
            self.sel_thresh = 1
            self.ratio_thresh = .5

        elif method == 'Concordia':
            if self.frequency is None:
                self.frequency = (10, 16)
            self.duration = (0.5, 3)
            self.det_butter = {'order': 2,
                               'freq': self.frequency,
                               }
            self.moving_rms = {'dur': .2,
                               'step': None}
            self.smooth = {'dur': .2,
                           'win': 'flat'}
            self.det_thresh = 3
            self.det_thresh_hi = 10
            self.tolerance = 0.2
            self.sel_thresh = 1

        else:
            raise ValueError('Unknown method')

        if frequency is not None:
            self.frequency = frequency

        if duration is not None:
            self.duration = duration

    def __repr__(self):
        return ('detsp_{0}_{1:02}-{2:02}Hz_{3:04.1f}-{4:04.1f}s'
                ''.format(self.method, self.frequency[0], self.frequency[1],
                          self.duration[0], self.duration[1]))

    def __call__(self, data, parent=None):
        """Detect spindles on the data.

        Parameters
        ----------
        data : instance of Data
            data used for detection
        parent : QWidget
            for use with GUI, as parent widget for the progress bar

        Returns
        -------
        instance of graphoelement.Spindles
            description of the detected spindles
        """
        if parent is not None:
            progress = QProgressDialog('Finding spindles', 'Abort',
                                       0, data.number_of('chan')[0], parent)
            progress.setWindowModality(Qt.ApplicationModal)

        spindle = Spindles()
        spindle.chan_name = data.axis['chan'][0]
        spindle.det_values = empty(data.number_of('chan')[0], dtype='O')
        spindle.density = zeros(data.number_of('chan')[0])

        if self.duration[1] is None:
            self.duration = self.duration[0], MAX_DURATION

        all_spindles = []
        i = 0
        for i, chan in enumerate(data.axis['chan'][0]):

            lg.info('Detecting spindles on channel %s', chan)
            time = hstack(data.axis['time'])
            dat_orig = hstack(data(chan=chan))
            dat_orig = dat_orig - dat_orig.mean() # demean

            if self.method == 'Ferrarelli2007':
                sp_in_chan, values, density = detect_Ferrarelli2007(dat_orig,
                                                                    data.s_freq,
                                                                    time,
                                                                    self)

            elif self.method == 'Moelle2011':
                sp_in_chan, values, density = detect_Moelle2011(dat_orig,
                                                                data.s_freq,
                                                                time, self)

            elif self.method == 'Nir2011':
                sp_in_chan, values, density = detect_Nir2011(dat_orig,
                                                             data.s_freq,
                                                             time, self)


            elif self.method == 'Wamsley2012':
                sp_in_chan, values, density = detect_Wamsley2012(dat_orig,
                                                                 data.s_freq,
                                                                 time, self)

            elif self.method == 'Martin2013':
                sp_in_chan, values, density = detect_Martin2013(dat_orig,
                                                                data.s_freq,
                                                                time, self)

            elif self.method == 'Ray2015':
                sp_in_chan, values, density = detect_Ray2015(dat_orig,
                                                            data.s_freq,
                                                            time, self)

            elif self.method == 'Lacourse2018':
                sp_in_chan, values, density = detect_Lacourse2018(dat_orig,
                                                                  data.s_freq,
                                                                  time, self)

            elif self.method == 'FASST':
                sp_in_chan, values, density = detect_FASST(dat_orig,
                                                           data.s_freq,
                                                           time, self,
                                                           submethod='abs')

            elif self.method == 'FASST2':
                sp_in_chan, values, density = detect_FASST(dat_orig,
                                                           data.s_freq,
                                                           time, self,
                                                           submethod='rms')

            elif self.method == 'UCSD':
                sp_in_chan, values, density = detect_UCSD(dat_orig,
                                                          data.s_freq, time,
                                                          self)

            elif self.method == 'Concordia':
                sp_in_chan, values, density = detect_Concordia(dat_orig,
                                                               data.s_freq,
                                                               time, self)

            else:
                raise ValueError('Unknown method')

            spindle.det_values[i] = values
            spindle.density[i] = density

            for sp in sp_in_chan:
                sp.update({'chan': chan})

            all_spindles.extend(sp_in_chan)

            if parent is not None:
                progress.setValue(i)
                if progress.wasCanceled():
                    return
            # end of loop over chan

        spindle.events = sorted(all_spindles, key=lambda x: x['start'])
        lg.info(str(len(spindle.events)) + ' spindles detected.')

        if self.merge and len(data.axis['chan'][0]) > 1:
            spindle.events = merge_close(spindle.events, self.min_interval)

        if parent is not None:
            progress.setValue(i + 1)

        return spindle


def detect_Ferrarelli2007(dat_orig, s_freq, time, opts):
    """Spindle detection based on Ferrarelli et al. 2007, and scripts obtained
    from Warby et al. (2014).

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_remez' : dict
            parameters for 'remez',
        'det_thresh' : float
            detection threshold
        'sel_thresh' : float
            selection threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' with nan,
        'sel_value' with selection value
    float
        spindle density, per 30-s epoch

    References
    ----------
    Ferrarelli, F. et al. Am. J. Psychiatry 164, 483-92 (2007).
    Warby, S. C. et al. Nat. Meth. 11(4), 385-92 (2014).
    """
    dat_det = transform_signal(dat_orig, s_freq, 'remez', opts.det_remez)
    dat_det = transform_signal(dat_det, s_freq, 'abs')

    idx_env = peaks_in_time(dat_det)
    envelope = dat_det[idx_env]
    idx_peak = idx_env[peaks_in_time(envelope)] # in raw data time
    idx_trough = peaks_in_time(envelope, troughs=True) # in envelope time
    troughs = ones(len(envelope)) * -1
    troughs[idx_trough] = envelope[idx_trough] # all non-trough values are -1

    det_value = define_threshold(dat_det, s_freq, 'mean', opts.det_thresh)
    sel_value = define_threshold(dat_det[idx_peak], s_freq, 'histmax',
                                 opts.sel_thresh, nbins=120)

    events_env = detect_events(envelope, 'above_thresh', det_value)

    if events_env is not None:
        events_env = _merge_close(envelope, events_env, time[idx_env],
                                  opts.tolerance)
        events_env = select_events(troughs, events_env,
                                   'Ferrarelli2007', sel_value)
        events = idx_env[events_env]
        # merging is necessary, because detected spindles may overlap if the
        # signal envelope does not dip below sel_thresh between two peaks above
        # det_thresh
        events = _merge_close(dat_det, events, time, opts.min_interval)
        events = within_duration(events, time, opts.duration)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_det,
                                   dat_orig, time, s_freq)
        lg.info('Spindles in chan: ' + str(len(sp_in_chan)))

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value, 'sel_value': sel_value}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Moelle2011(dat_orig, s_freq, time, opts):
    """Spindle detection based on Moelle et al. 2011

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_remez' : dict
            parameters for 'remez',
        'moving_rms' : dict
            parameters for 'moving_rms'
        'smooth' : dict
            parameters for 'smooth'
        'det_thresh' : float
            detection threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' with nan,
        'sel_value' with nan
    float
        spindle density, per 30-s epoch

    References
    ----------
    Moelle, M. et al. J. Neurosci. 22(24), 10941-7 (2002).
    """
    dat_det = transform_signal(dat_orig, s_freq, 'remez', opts.det_remez)
    dat_det = transform_signal(dat_det, s_freq, 'moving_rms', opts.moving_rms)
    dat_det = transform_signal(dat_det, s_freq, 'smooth', opts.smooth)

    det_value = define_threshold(dat_det, s_freq, 'mean+std',
                                 opts.det_thresh)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = _merge_close(dat_det, events, time, opts.tolerance)
        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_det, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_det,
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value, 'sel_value': nan}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Nir2011(dat_orig, s_freq, time, opts):
    """Spindle detection based on Nir et al. 2011

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_butter' : dict
            parameters for 'butter',
        'smooth' : dict
            parameters for 'gaussian'
        'det_thresh' : float
            detection threshold
        'sel_thresh' : float
            selection threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' with nan,
        'sel_value' with selection value
    float
        spindle density, per 30-s epoch

    Notes
    -----
    This paper also selects channels carefully:
    'First, the channels with spindle activity in NREM sleep were
    chosen for further analysis.'

    'Third, those channels, in which an increase in spectral power
    within the detected events was restricted to the spindle-frequency
    range (10-16 Hz) rather than broadband.'

    References
    ----------
    Nir, Y. et al. Neuron 70, 153-69 (2011).
    """
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)
    dat_det = transform_signal(dat_det, s_freq, 'hilbert')
    dat_det = transform_signal(dat_det, s_freq, 'abs')
    dat_det = transform_signal(dat_det, s_freq, 'gaussian', opts.smooth)

    det_value = define_threshold(dat_det, s_freq, 'mean+std',
                                 opts.det_thresh)
    sel_value = define_threshold(dat_det, s_freq, 'mean+std', opts.sel_thresh)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = _merge_close(dat_det, events, time, opts.tolerance)
        events = select_events(dat_det, events, 'above_thresh', sel_value)

        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_det, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_det,
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value, 'sel_value': sel_value}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Wamsley2012(dat_orig, s_freq, time, opts):
    """Spindle detection based on Wamsley et al. 2012

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_wavelet' : dict
            parameters for 'morlet',
        'smooth' : dict
            parameters for 'smooth'
        'det_thresh' : float
            detection threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' is nan,
        'sel_value' is nan (for consistency with other methods)
    float
        spindle density, per 30-s epoch

    References
    ----------
    Wamsley, E. J. et al. Biol. Psychiatry 71, 154-61 (2012).
    """
    dat_wav = transform_signal(dat_orig, s_freq, 'morlet', opts.det_wavelet)
    dat_det = real(dat_wav ** 2) ** 2
    dat_det = transform_signal(dat_det, s_freq, 'smooth', opts.smooth)

    det_value = define_threshold(dat_det, s_freq, 'mean', opts.det_thresh)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = _merge_close(dat_det, events, time, opts.tolerance)
        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_det, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers,
                                   absolute(dat_wav), dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value, 'sel_value': nan}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Martin2013(dat_orig, s_freq, time, opts):
    """Spindle detection based on Martin et al. 2013

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_remez' : dict
             parameters for 'remez' filter
        'moving_rms' : dict
             parameters for 'moving_rms'
        'det_thresh' : float
            percentile for detection threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' is nan,
        'sel_value' is nan (for consistency with other methods)
    float
        spindle density, per 30-s epoch

    References
    ----------
    Martin, N. et al. Neurobio Aging 34(2), 468-76 (2013).
    """
    dat_filt = transform_signal(dat_orig, s_freq, 'remez', opts.det_remez)
    dat_det = transform_signal(dat_filt, s_freq, 'moving_rms', opts.moving_rms)
        # downsampled

    det_value = percentile(dat_det, opts.det_thresh)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events *= int(around(s_freq * opts.moving_rms['step'])) # upsample
        events = _merge_close(dat_filt, events, time, opts.tolerance)
        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_filt, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_filt,
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value, 'sel_value': nan}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Ray2015(dat_orig, s_freq, time, opts):
    """Spindle detection based on Ray et al., 2015

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'cdemod' : dict
            parameters for 'cdemod' (complex demodulation)
        'det_butter' : dict
            parameters for 'butter',
        'det_low_butter' : dict
            parameters for 'low_butter',
        'smooth' : dict
            parameters for 'smooth'
        'zscore' : dict
            parameters for 'moving_zscore'
        'det_thresh' : float
            detection threshold
        'sel_thresh' : nan
            selection threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' is nan,
        'sel_value' is nan (for consistency with other methods)
    float
        spindle density, per 30-s epoch

    References
    ----------
    Ray, L. B. et al. Front. Hum. Neurosci. 9-16 (2015).
    """
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)
    dat_det = transform_signal(dat_det, s_freq, 'cdemod', opts.cdemod)
    dat_det = transform_signal(dat_det, s_freq, 'low_butter',
                               opts.det_low_butter)
    dat_det = transform_signal(dat_det, s_freq, 'smooth', opts.smooth)
    dat_det = transform_signal(dat_det, s_freq, 'abs_complex')
    dat_det = transform_signal(dat_det, s_freq, 'moving_zscore', opts.zscore)

    det_value = opts.det_thresh
    sel_value = opts.sel_thresh

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = _merge_close(dat_det, events, time, opts.tolerance)
        events = select_events(dat_det, events, 'above_thresh', sel_value)

        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_det, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_det,
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value, 'sel_value': sel_value}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Lacourse2018(dat_orig, s_freq, time, opts):
    """Spindle detection based on Lacourse et al., 2018

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_butter' : dict
            parameters for 'butter',
        'det_butter2' : dict
            parameters for 'butter',
        'windowing' :
            'step' for downsampling and 'dur' for moving window duration
        'moving_ms' : dict
            parameters for 'moving_rms'
        'moving_power_ratio' :
            parameters for 'moving_power_ratio'
        'zscore' :
            parameters for 'moving_zscore'
        'moving_covar' :
            parameters for 'moving_covar'
        'moving_sd' :
            parameters for 'moving_sd'
        'smooth' : dict
            parameters for 'smooth'
        'abs_pow_thresh' : float
            absolute power threshold
        'rel_pow_thresh' : float
            relative power threshold
        'covar_thresh' : float
            covariance threshold
        'corr_thresh' : float
            coorelation threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' is nan,
        'sel_value' is nan (for consistency with other methods)
    float
        spindle density, per 30-s epoch

    References
    ----------
    Lacourse, K. et al. J. Neurosci. Meth. (2018).
    """
    # Prepare downsampling
    step = opts.windowing['step']
    if step:
        ds_freq = int(1 / step) # downsampled sampling frequency
        opts.tolerance *= step
    else:
        ds_freq = s_freq

    # Absolute sigma power
    dat_sigma = transform_signal(dat_orig, s_freq, 'double_sosbutter',
                                 opts.det_butter)
    dat_det = transform_signal(dat_sigma, s_freq, 'moving_ms', opts.moving_ms)
    dat_det[dat_det <= 0] = 0.000000001 # arbitrarily small value
    abs_sig_pow = log10(dat_det)
        # Option to adapt the absolute threshold, for low-amplitude recordings
    if opts.abs_pow_thresh < 0:
        opts.abs_pow_thresh = (mean(abs_sig_pow) -
                               opts.abs_pow_thresh * std(abs_sig_pow))
    abs_sig_pow = transform_signal(abs_sig_pow, ds_freq, 'smooth', opts.smooth)

    # Relative sigma power
    dat_det = transform_signal(dat_orig, s_freq, 'moving_power_ratio',
                               opts.moving_power_ratio)
    dat_det[dat_det <= 0] = 0.000000001
    dat_det = log10(dat_det)
    rel_sig_pow = transform_signal(dat_det, ds_freq, 'moving_zscore',
                                   opts.zscore)
    rel_sig_pow = transform_signal(rel_sig_pow, ds_freq, 'smooth', opts.smooth)

    # Sigma covariance
    dat_broad = transform_signal(dat_orig, s_freq, 'double_sosbutter',
                                 opts.det_butter2)
    dat_covar = transform_signal(dat_sigma, s_freq, 'moving_covar',
                                 opts.moving_covar, dat2=dat_broad)
    dat_det = dat_covar.copy()
    dat_det[dat_det < 0] = 0 # negative covariances are discarded
    dat_det = log10(dat_det + 1) # add 1 to avoid -inf
    sigma_covar = transform_signal(dat_det, ds_freq, 'moving_zscore',
                                   opts.zscore)
    sigma_covar = transform_signal(sigma_covar, ds_freq, 'smooth', opts.smooth)

    # Sigma correlation
    dat_sd_broad = transform_signal(dat_broad, s_freq, 'moving_sd',
                                    opts.moving_sd)
    dat_sd_sigma = transform_signal(dat_sigma, s_freq, 'moving_sd',
                                    opts.moving_sd)
    dat_sd_broad[dat_sd_broad == 0] = 0.000000001
    dat_sd_sigma[dat_sd_sigma == 0] = 0.000000001
    sigma_corr = dat_covar / (dat_sd_broad * dat_sd_sigma)
    sigma_corr = transform_signal(sigma_corr, ds_freq, 'smooth', opts.smooth)

    # Thresholding
    abs_and_cov = logical_and(abs_sig_pow >= opts.abs_pow_thresh,
                              sigma_covar >= opts.covar_thresh)
    concensus = logical_and.reduce((rel_sig_pow >= opts.rel_pow_thresh,
                                    sigma_corr >= opts.corr_thresh,
                                    abs_and_cov))
    events = detect_events(concensus, 'custom') # at s_freq * 0.1

    if events is not None:
        events = _merge_close(dat_sigma, events, time, opts.tolerance)
        events = _select_period(events, abs_and_cov) + 1

        if step:
            events = events * (s_freq * step) # upsample
            events = asarray(around(events), dtype=int)

        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_sigma, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_sigma,
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'abs_pow_thresh': opts.abs_pow_thresh,
              'rel_pow_thresh': opts.rel_pow_thresh,
              'covar_thresh': opts.covar_thresh,
              'corr_thresh': opts.corr_thresh}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_FASST(dat_orig, s_freq, time, opts, submethod='rms'):
    """Spindle detection based on FASST method, itself based on Moelle et al.
    (2002).

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        'det_butter' : dict
            parameters for 'butter',
        'moving_rms' : dict
            parameters for 'moving_rms'
        'smooth' : dict
            parameters for 'smooth'
        'det_thresh' : float
            detection threshold (percentile)
    submethod : str
        'abs' (rectified) or 'rms' (root-mean-square)

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' with nan,
        'sel_value' with nan
    float
        spindle density, per 30-s epoch

    References
    ----------
    Leclercq, Y. et al. Compu. Intel. and Neurosci. (2011).
    """
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)

    det_value = percentile(dat_det, opts.det_thresh)

    if submethod == 'abs':
        dat_det = transform_signal(dat_det, s_freq, 'abs')
    elif submethod == 'rms':
        dat_det = transform_signal(dat_det, s_freq, 'moving_rms',
                                   opts.moving_rms)

    dat_det = transform_signal(dat_det, s_freq, 'smooth', opts.smooth)

    events = detect_events(dat_det, 'above_thresh', det_value)

    if events is not None:
        events = _merge_close(dat_det, events, time, opts.tolerance)
        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_det, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_det,
                               dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value, 'sel_value': nan}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_UCSD(dat_orig, s_freq, time, opts):
    """Spindle detection based on the UCSD method

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSpindle'
        det_wavelet : dict
            parameters for 'wavelet_real',
        det_thres' : float
            detection threshold
        sel_wavelet : dict
            parameters for 'wavelet_real'
        sel_thresh : float
            selection threshold
        ratio_thresh : float
            ratio between power inside and outside spindle band to accept them

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo' with detection value, 'det_value_hi' with nan,
        'sel_value' with selection value
    float
        spindle density, per 30-s epoch

    """
    dat_det = transform_signal(dat_orig, s_freq, 'wavelet_real',
                               opts.det_wavelet)

    det_value = define_threshold(dat_det, s_freq, 'median+std',
                                 opts.det_thresh)

    events = detect_events(dat_det, 'maxima', det_value)

    dat_sel = transform_signal(dat_orig, s_freq, 'wavelet_real',
                               opts.sel_wavelet)
    sel_value = define_threshold(dat_sel, s_freq, 'median+std',
                                 opts.sel_thresh)
    events = select_events(dat_sel, events, 'above_thresh', sel_value)

    events = _merge_close(dat_det, events, time, opts.tolerance)
    events = within_duration(events, time, opts.duration)
    events = _merge_close(dat_det, events, time, opts.min_interval)
    events = remove_straddlers(events, time, s_freq)

    events = power_ratio(events, dat_orig, s_freq, opts.frequency,
                         opts.ratio_thresh)

    power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
    powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
    sp_in_chan = make_spindles(events, power_peaks, powers, dat_det,
                               dat_orig, time, s_freq)

    values = {'det_value_lo': det_value, 'sel_value': sel_value}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def detect_Concordia(dat_orig, s_freq, time, opts):
    """Spindle detection, experimental Concordia method. Similar to Moelle 2011
    and Nir2011.

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    opts : instance of 'DetectSpindle'
        'det_butter' : dict
            parameters for 'butter',
        'moving_rms' : dict
            parameters for 'moving_rms'
        'smooth' : dict
            parameters for 'smooth'
        'det_thresh' : float
            low detection threshold
        'det_thresh_hi' : float
            high detection threshold
        'sel_thresh' : float
            selection threshold

    Returns
    -------
    list of dict
        list of detected spindles
    dict
        'det_value_lo', 'det_value_hi' with detection values, 'sel_value' with
        selection value
    float
        spindle density, per 30-s epoch
    """
    dat_det = transform_signal(dat_orig, s_freq, 'butter', opts.det_butter)
    dat_det = transform_signal(dat_det, s_freq, 'moving_rms', opts.moving_rms)
    dat_det = transform_signal(dat_det, s_freq, 'smooth', opts.smooth)

    det_value_lo = define_threshold(dat_det, s_freq, 'mean+std',
                                    opts.det_thresh)
    det_value_hi = define_threshold(dat_det, s_freq, 'mean+std',
                                    opts.det_thresh_hi)
    sel_value = define_threshold(dat_det, s_freq, 'mean+std', opts.sel_thresh)

    events = detect_events(dat_det, 'between_thresh',
                           value=(det_value_lo, det_value_hi))

    if events is not None:
        events = _merge_close(dat_det, events, time, opts.tolerance)

        events = select_events(dat_det, events, 'above_thresh', sel_value)

        events = within_duration(events, time, opts.duration)
        events = _merge_close(dat_det, events, time, opts.min_interval)
        events = remove_straddlers(events, time, s_freq)

        power_peaks = peak_in_power(events, dat_orig, s_freq, opts.power_peaks)
        powers = power_in_band(events, dat_orig, s_freq, opts.frequency)
        sp_in_chan = make_spindles(events, power_peaks, powers, dat_det,
                                   dat_orig, time, s_freq)

    else:
        lg.info('No spindle found')
        sp_in_chan = []

    values = {'det_value_lo': det_value_lo, 'sel_value': sel_value}

    density = len(sp_in_chan) * s_freq * 30 / len(dat_orig)

    return sp_in_chan, values, density


def transform_signal(dat, s_freq, method, method_opt=None, dat2=None):
    """Transform the data using different methods.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with all the data for one channel
    s_freq : float
        sampling frequency
    method : str
        one of 'abs', 'abs_complex', 'butter', 'cdemod', 'cheby2',
        'double_butter', 'double_sosbutter', 'gaussian', 'hilbert',
        'high_butter', 'low_butter', 'morlet', 'moving_covar', 'moving_ms',
        'moving_periodogram', 'moving_power_ratio',  'moving_rms', 'moving_sd',
        'moving_zscore', 'remez', 'smooth', 'sosbutter', 'spectrogram',
        'wavelet_real'.
    method_opt : dict
        depends on methods
    dat2 : ndarray(dtype='float')
        second vector with data

    Returns
    -------
    ndarray (dtype='float')
        vector with all the data for one channel

    Notes
    -----
    double_butter implements an effective bandpass by applying a highpass,
    followed by a lowpass. This method reduces filter instability, due to
    underlying numerical instability arising from nyquist / freq at low freq.

    Wavelets pass only absolute values already, it does not make sense to store
    the complex values.

    Methods
    -------
    butter has parameters:
        freq : tuple of float
            low and high values for bandpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    cdemod has parameters:
        freq : float
            carrier frequency for complex demodulation

    cheby2 has parameters:
        freq : tuple of float
            low and high values for bandpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    double_butter has parameters:
        freq : tuple of float
            low and high values for highpass, then lowpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    double_sosbutter has parameters:
        freq : tuple of float
            low and high values for highpass, then lowpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    gaussian has parameters:
        dur : float
            standard deviation of the Gaussian kernel, aka sigma (sec)

    high_butter has parameters:
        freq : float
            Highpass (lowcut) frequency, in Hz
        order : int
            filter order (will be effecively doubled by filtfilt)

    low_butter has parameters:
        freq : float
            Lowpass (highcut) frequency, in Hz
        order : int
            filter order (will be effecively doubled by filtfilt)

    morlet has parameters:
        f0 : float
            center frequency in Hz
        sd : float
            standard deviation of frequency
        dur : float
            window length in number of standard deviations

    moving_covar has parameters:
        dur : float
            duration of the window (sec)
        step: float
            step between consecutive windows (sec)

    moving_ms has parameters:
        dur : float
            duration of the window (sec)
        step: float
            step between consecutive windows (sec)

    moving_periodogram has parameters:
        dur : float
            duration of the z-score sliding window (sec)
        freq : tuple of float
            frequency range for periodogram (Hz)
        step: float
            step between consecutive windows (sec)

    moving_power_ratio has parameters:
        dur : float
            duration of the z-score sliding window (sec)
        freq_narrow : tuple of float
            frequency range for the narrowband power (Hz)
        freq_broad : tuple of float
            frequency range for the broadband power (Hz)
        fft_dur : float
            duration of the FFT window (sec)
        step: float
            step between consecutive windows (sec)

    moving_rms has parameters:
        dur : float
            duration of the window (sec)
        step: float
            step between consecutive windows (sec)

    moving_sd has parameters:
        dur : float
            duration of the z-score sliding window (sec)
        step: float
            step between consecutive windows (sec)

    moving_zscore has parameters:
        dur : float
            duration of the z-score sliding window (sec)
        pcl_range : tuple of float, or None
            if not None, only data within this percentile range will be used
            for determining the standard deviation for calculation of the
            z-score
        step: float
            step between consecutive windows (sec)

    remez has parameters:
        freq : tuple of float
            low and high values for bandpass
        rolloff : float
            bandwidth, in hertz, between stop and pass frequencies
        dur : float
            dur * s_freq = N, where N is the filter order, a.k.a number of taps

    smooth has parameters:
        dur : float
            duration of the convolution window (sec). For 'triangle', base of
            isosceles triangle.

    wavelet_real has parameters:
        freqs : ndarray
            vector of wavelet frequencies for spindle detection
        dur : float
            duration of the wavelet (sec)
        width : float
            wavelet width
        win : float
            moving average window length (sec) of wavelet convolution
    """
    if 'abs' == method:
        dat = absolute(dat)

    if 'abs_complex' == method:
        dat = dat.real**2 + dat.imag**2

    if 'butter' == method:
        freq = method_opt['freq']
        N = method_opt['order']

        nyquist = s_freq / 2
        Wn = asarray(freq) / nyquist
        b, a = butter(N, Wn, btype='bandpass')
        dat = filtfilt(b, a, dat)

    if 'cdemod' == method:
        carr_freq = method_opt['freq']

        carr_sig = exp(-1j * 2 * pi * carr_freq * arange(0, len(dat)) / s_freq)
        dat = dat * carr_sig

    if 'cheby2' == method:
        freq = method_opt['freq']
        N = method_opt['order']

        Rs = 40
        nyquist = s_freq / 2
        Wn = asarray(freq) / nyquist
        b, a = cheby2(N, Rs, Wn, btype='bandpass')
        dat = filtfilt(b, a, dat)

    if 'double_butter' == method:
        freq = method_opt['freq']
        N = method_opt['order']
        nyquist = s_freq / 2

        # Highpass
        Wn = freq[0] / nyquist
        b, a = butter(N, Wn, btype='highpass')
        dat = filtfilt(b, a, dat)

        # Lowpass
        Wn = freq[1] / nyquist
        b, a = butter(N, Wn, btype='lowpass')
        dat = filtfilt(b, a, dat)

    if 'double_sosbutter' == method:
        freq = method_opt['freq']
        N = method_opt['order']
        nyquist = s_freq / 2

        # Highpass
        Wn = freq[0] / nyquist
        sos = butter(N, Wn, btype='highpass', output='sos')
        dat = sosfiltfilt(sos, dat)

        # Lowpass
        Wn = freq[1] / nyquist
        sos = butter(N, Wn, btype='lowpass', output='sos')
        dat = sosfiltfilt(sos, dat)

    if 'gaussian' == method:
        sigma = method_opt['dur']

        dat = gaussian_filter(dat, sigma)

    if 'hilbert' == method:
        N = len(dat)
        dat = hilbert(dat, N=next_fast_len(N)) # much faster this way
        dat = dat[:N] # truncate away zero-padding


    if 'high_butter' == method:
        freq = method_opt['freq']
        N = method_opt['order']

        nyquist = s_freq / 2
        Wn = freq / nyquist
        b, a = butter(N, Wn, btype='highpass')
        dat = filtfilt(b, a, dat)

    if 'low_butter' == method:
        freq = method_opt['freq']
        N = method_opt['order']

        nyquist = s_freq / 2
        Wn = freq / nyquist
        b, a = butter(N, Wn, btype='lowpass')
        dat = filtfilt(b, a, dat)

    if 'morlet' == method:
        f0 = method_opt['f0']
        sd = method_opt['sd']
        dur = method_opt['dur']
        output = method_opt['output']

        wm = _wmorlet(f0, sd, s_freq, dur)
        dat = fftconvolve(dat, wm, mode='same')
        if 'absolute' == output:
            dat = absolute(dat)

    if 'moving' in method:
        dur = method_opt['dur']
        halfdur = dur / 2
        total_dur = len(dat) / s_freq
        last = len(dat) - 1

        if method_opt['step']:
            step = method_opt['step']
            len_out = int(len(dat) / (step * s_freq))
        else:
            step = 1 / s_freq
            len_out = len(dat)

        out = zeros((len_out))

        if 'moving_covar' == method:
            for i, j in enumerate(arange(0, total_dur, step)[:-1]):
                beg = max(0, int((j - halfdur) * s_freq))
                end = min(last, int((j + halfdur) * s_freq))
                win1 = dat[beg:end]
                win2 = dat2[beg:end]
                out[i] = mean((win1 - mean(win1)) * (win2 - mean(win2)))
            dat = out

        if 'moving_periodogram' == method:
            nfft = next_fast_len(dur * s_freq)
            sf = rfftfreq(nfft, 1 / s_freq)
            freq = method_opt['freq']
            f0 = asarray([abs(x - freq[0]) for x in sf]).argmin()
            f1 = asarray([abs(x - freq[1]) for x in sf]).argmin()
            out = zeros((len_out, f1 - f0))

            for i, j in enumerate(arange(0, total_dur, step)[:-1]):
                beg = max(0, int((j - halfdur) * s_freq))
                end = min(last, int((j + halfdur) * s_freq))
                windat = dat[beg:end]
                sf, psd = periodogram(windat, s_freq, 'hann', nfft=nfft,
                                       detrend='constant')
                out[i, :] = psd[f0:f1]

            dat = out

        if 'moving_power_ratio' == method:
            freq1 = method_opt['freq_narrow']
            freq2 = method_opt['freq_broad']
            fft_dur = method_opt['fft_dur']
            nfft = int(s_freq * fft_dur)

            for i, j in enumerate(arange(0, total_dur, step)[:-1]):
                beg = max(0, int((j - halfdur) * s_freq))
                end = min(last, int((j + halfdur) * s_freq))
                windat = dat[beg:end]

                sf, psd = periodogram(windat, s_freq, 'hann', nfft=nfft,
                                       detrend='constant')
                f0 = asarray([abs(x - freq1[0]) for x in sf]).argmin()
                f1 = asarray([abs(x - freq1[1]) for x in sf]).argmin()
                pow1 = sum(psd[f0:f1])

                f0 = asarray([abs(x - freq2[0]) for x in sf]).argmin()
                f1 = asarray([abs(x - freq2[1]) for x in sf]).argmin()
                pow2 = sum(psd[f0:f1])

                out[i] = pow1 / pow2

            dat = out

        if 'moving_sd' == method:
            for i, j in enumerate(arange(0, total_dur, step)[:-1]):
                beg = max(0, int((j - halfdur) * s_freq))
                end = min(last, int((j + halfdur) * s_freq))
                win = dat[beg:end]
                out[i] = std(win)
            dat = out

        if 'moving_zscore' == method:
            pcl_range = method_opt['pcl_range']
            if pcl_range is not None:
                lo = percentile(dat, pcl_range[0])
                hi = percentile(dat, pcl_range[1])

            for i, j in enumerate(arange(0, total_dur, step)[:-1]):
                beg = max(0, int((j - halfdur) * s_freq))
                end = min(last, int((j + halfdur) * s_freq))
                windat = stddat = dat[beg:end]

                if pcl_range is not None:
                    stddat = windat[logical_and(windat > lo, windat < hi)]
                out[i] = (dat[i] - mean(windat)) / std(stddat)
            dat = out

        if method in ['moving_rms', 'moving_ms']:
            for i, j in enumerate(arange(0, total_dur, step)[:-1]):
                beg = max(0, int((j - halfdur) * s_freq))
                end = min(last, int((j + halfdur) * s_freq))
                out[i] = mean(square(dat[beg:end]))
            if method == 'moving_rms':
                out = sqrt(out)
            dat = out

    if 'remez' == method:
        Fp1, Fp2 = method_opt['freq']
        rolloff = method_opt['rolloff']
        dur = method_opt['dur']

        N = int(s_freq * dur)
        nyquist = s_freq / 2
        Fs1, Fs2 = Fp1 - rolloff, Fp2 + rolloff
        dens = 20
        bpass = remez(N, [0, Fs1, Fp1, Fp2, Fs2, nyquist], [0, 1, 0],
                      grid_density=dens, fs=s_freq)
        dat = filtfilt(bpass, 1, dat)

    if 'smooth' == method:
        dur = method_opt['dur']
        win = method_opt['win']

        if 'flat' in win:
            flat = ones(int(dur * s_freq))
            H = flat / sum(flat)

            if 'flat_left' == win:
                H = concatenate((H, zeros(len(H))))
            elif 'flat_right' == win:
                H = concatenate((zeros(len(H) - 1), H))

        elif 'triangle' == win:
            T = int(dur * s_freq / 2)
            a = arange(T, 0, -1)

            H = hstack([a[-1:0:-1], a])
            H = H / sum(H)

        dat = fftconvolve(dat, H, mode='same')

    if 'sosbutter' == method:
        freq = method_opt['freq']
        N = method_opt['order']

        nyquist = s_freq / 2
        Wn = asarray(freq) / nyquist
        sos = butter(N, Wn, btype='bandpass', output='sos')
        dat = sosfiltfilt(sos, dat)

    if 'spectrogram' == method:
        nperseg = method_opt['dur'] * s_freq
        noverlap = method_opt['step'] * s_freq
        detrend = method_opt['detrend']

        dat = spectrogram(dat, fs=s_freq, nperseg=nperseg, noverlap=noverlap,
                          detrend=detrend)

    if 'wavelet_real' == method:
        freqs = method_opt['freqs']
        dur = method_opt['dur']
        width = method_opt['width']
        win = int(method_opt['win'] * s_freq)

        wm = _realwavelets(s_freq, freqs, dur, width)
        tfr = empty((dat.shape[0], wm.shape[0]))
        for i, one_wm in enumerate(wm):
            x = abs(fftconvolve(dat, one_wm, mode='same'))
            tfr[:, i] = fftconvolve(x, tukey(win), mode='same')
        dat = mean(tfr, axis=1)

    return dat


def define_threshold(dat, s_freq, method, value, nbins=120):
    """Return the value of the threshold based on relative values.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after selection-transformation
    s_freq : float
        sampling frequency
    method : str
        one of 'mean', 'median', 'std', 'mean+std', 'median+std', 'histmax'
    value : float
        value to multiply the values for
    nbins : int
        for histmax method only, number of bins in the histogram

    Returns
    -------
    float
        threshold in useful units.

    """
    if method == 'mean':
        value = value * mean(dat)
    elif method == 'median':
        value = value * median(dat)
    elif method == 'std':
        value = value * std(dat)
    elif method == 'mean+std':
        value = mean(dat) + value * std(dat)
    elif method == 'median+std':
        value = median(dat) + value * std(dat)
    elif method == 'histmax':
        hist = histogram(dat, bins=nbins)
        idx_maxbin = argmax(hist[0])
        maxamp = mean((hist[1][idx_maxbin], hist[1][idx_maxbin + 1]))
        value = value * maxamp

    return value


def peaks_in_time(dat, troughs=False):
    """Find indices of peaks or troughs in data.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data
    troughs : bool
        if True, will return indices of troughs instead of peaks

    Returns
    -------
    nadarray of int
        indices of peaks (or troughs) in dat

    Note
    ----
    This function does not deal well with flat signal; when the signal is not
    increasing, it is assumed to be descreasing. As a result, this function
    finds troughs where the signal begins to increase after either decreasing
    or remaining constant
    """
    diff_dat = diff(dat)
    increasing = zeros(len(diff_dat))
    increasing[diff_dat > 0] = 1 # mask for all points where dat is increasing
    flipping = diff(increasing) # peaks are -1, troughs are 1, the rest is zero

    target = -1 if not troughs else 1

    return where(flipping == target)[0] + 1


def detect_events(dat, method, value=None):
    """Detect events using 'above_thresh', 'below_thresh' or
    'maxima' method.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after transformation
    method : str
        'above_thresh', 'below_thresh' or 'maxima'
    value : float or tuple of float
        for 'above_thresh' or 'below_thresh', it's the value of threshold for
        the event detection
        for 'between_thresh', it's the lower and upper threshold as tuple
        for 'maxima', it's the distance in s from the peak to find a minimum

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    """
    if 'thresh' in method or 'custom' == method:

        if method == 'above_thresh':
            above_det = dat >= value
            detected = _detect_start_end(above_det)

        if method == 'below_thresh':
            below_det = dat < value
            detected = _detect_start_end(below_det)

        if method == 'between_thresh':
            above_det = dat >= value[0]
            below_det = dat < value[1]
            between_det = logical_and(above_det, below_det)
            detected = _detect_start_end(between_det)

        if method == 'custom':
            detected = _detect_start_end(dat)

        if detected is None:
            return None

        if method in ['above_thresh', 'custom']:
            # add the location of the peak in the middle
            detected = insert(detected, 1, 0, axis=1)
            for i in detected:
                i[1] = i[0] + argmax(dat[i[0]:i[2]])

        if method in ['below_thresh', 'between_thresh']:
            # add the location of the trough in the middle
            detected = insert(detected, 1, 0, axis=1)
            for i in detected:
                i[1] = i[0] + argmin(dat[i[0]:i[2]])

    if method == 'maxima':
        peaks = argrelmax(dat)[0]
        detected = vstack((peaks, peaks, peaks)).T

        if value is not None:
            detected = detected[dat[peaks] > value, :]

    return detected


def select_events(dat, detected, method, value):
    """Select start sample and end sample of the events.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after selection-transformation
    detected : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    method : str
        'above_thresh', 'below_thresh', 'below_thresh_positive'
    value : float
        for 'threshold', it's the value of threshold for the spindle selection.

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    """
    if method == 'above_thresh':
        above_sel = dat >= value
        detected = _select_period(detected, above_sel)
    elif method == 'below_thresh':
        below_sel = dat <= value
        detected = _select_period(detected, below_sel)
    elif method == 'Ferrarelli2007':
        below_sel = dat <= value
        positive = dat >= 0
        below_sel_positive = invert(logical_and(below_sel, positive))
        detected = _select_period(detected, below_sel_positive)

    return detected


def merge_close(events, min_interval, merge_to_longer=False):
    """Merge events that are separated by a less than a minimum interval.

    Parameters
    ----------
    events : list of dict
        events with 'start' and 'end' times, from one or several channels.
        **Events must be sorted by their start time.**
    min_interval : float
        minimum delay between consecutive events, in seconds
    merge_to_longer : bool (default: False)
        If True, info (chan, peak, etc.) from the longer of the 2 events is
        kept. Otherwise, info from the earlier onset spindle is kept.

    Returns
    -------
    list of dict
        original events list with close events merged.
    """
    half_iv = min_interval / 2
    merged = []

    for higher in events:

        if not merged:
            merged.append(higher)

        else:
            lower = merged[-1]

            if higher['start'] - half_iv <= lower['end'] + half_iv:

                if merge_to_longer and (higher['end'] - higher['start'] >
                lower['end'] - lower['start']):
                    start = min(lower['start'], higher['start'])
                    higher.update({'start': start})
                    merged[-1] = higher

                else:
                    end = max(lower['end'], higher['end'])
                    merged[-1].update({'end': end})

            else:
                merged.append(higher)

    return merged


def within_duration(events, time, limits):
    """Check whether event is within time limits.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x M matrix with start sample first and end samples last on M
    time : ndarray (dtype='float')
        vector with time points
    limits : tuple of float
        low and high limit for spindle duration

    Returns
    -------
    ndarray (dtype='int')
        N x M matrix with start sample first and end samples last on M
    """
    min_dur = max_dur = ones(events.shape[0], dtype=bool)

    if limits[0] is not None:
        min_dur = time[events[:, -1] - 1] - time[events[:, 0]] >= limits[0]

    if limits[1] is not None:
        max_dur = time[events[:, -1] - 1] - time[events[:, 0]] <= limits[1]

    return events[min_dur & max_dur, :]


def remove_straddlers(events, time, s_freq, tolerance=0.1):
    """Reject an event if it straddles a stitch, by comparing its
    duration to its timespan.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x M matrix with start, ..., end samples
    time : ndarray (dtype='float')
        vector with time points
    s_freq : float
        sampling frequency
    tolerance : float, def=0.1
        maximum tolerated difference between event duration and timespan

    Returns
    -------
    ndarray (dtype='int')
        N x M matrix with start , ..., end samples
    """
    dur = (events[:, -1] - 1 - events[:, 0]) / s_freq
    continuous = time[events[:, -1] - 1] - time[events[:, 0]] - dur < tolerance

    return events[continuous, :]



def power_ratio(events, dat, s_freq, limits, ratio_thresh):
    """Estimate the ratio in power between spindle band and lower frequencies.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the original data
    s_freq : float
        sampling frequency
    limits : tuple of float
        high and low frequencies for spindle band
    ratio_thresh : float
        ratio between spindle vs non-spindle amplitude

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    Notes
    -----
    In the original matlab script, it uses amplitude, not power.

    """
    ratio = empty(events.shape[0])
    for i, one_event in enumerate(events):

        x0 = one_event[0]
        x1 = one_event[2]

        if x0 < 0 or x1 >= len(dat):
            ratio[i] = 0

        else:
            f, Pxx = periodogram(dat[x0:x1], s_freq, scaling='spectrum')
            Pxx = sqrt(Pxx)  # use amplitude

            freq_sp = (f >= limits[0]) & (f <= limits[1])
            freq_nonsp = (f <= limits[1])

            ratio[i] = mean(Pxx[freq_sp]) / mean(Pxx[freq_nonsp])

    events = events[ratio > ratio_thresh, :]

    return events


def peak_in_power(events, dat, s_freq, method, value=None):
    """Define peak in power of the signal.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the original data
    s_freq : float
        sampling frequency
    method : str or None
        'peak' or 'interval'. If None, values will be all NaN
    value : float
        size of the window around peak, or nothing (for 'interval')

    Returns
    -------
    ndarray (dtype='float')
        vector with peak frequency

    """
    dat = diff(dat)  # remove 1/f

    peak = empty(events.shape[0])
    peak.fill(nan)

    if method is not None:
        for i, one_event in enumerate(events):

            if method == 'peak':
                x0 = one_event[1] - value / 2 * s_freq
                x1 = one_event[1] + value / 2 * s_freq

            elif method == 'interval':
                x0 = one_event[0]
                x1 = one_event[2]

            if x0 < 0 or x1 >= len(dat):
                peak[i] = nan
            else:
                f, Pxx = periodogram(dat[x0:x1], s_freq)
                idx_peak = Pxx[f < MAX_FREQUENCY_OF_INTEREST].argmax()
                peak[i] = f[idx_peak]

    return peak


def power_in_band(events, dat, s_freq, frequency):
    """Define power of the signal within frequency band.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the original data
    s_freq : float
        sampling frequency
    frequency : tuple of float
        low and high frequency of spindle band, for window

    Returns
    -------
    ndarray (dtype='float')
        vector with power
    """
    dat = diff(dat)  # remove 1/f

    pw = empty(events.shape[0])
    pw.fill(nan)

    for i, one_event in enumerate(events):

        x0 = one_event[0]
        x1 = one_event[2]

        if x0 < 0 or x1 >= len(dat):
            pw[i] = nan
        else:
            sf, Pxx = periodogram(dat[x0:x1], s_freq)
            # find nearest frequencies in sf
            b0 = asarray([abs(x - frequency[0]) for x in sf]).argmin()
            b1 = asarray([abs(x - frequency[1]) for x in sf]).argmin()
            pw[i] = mean(Pxx[b0:b1])

    return pw


def make_spindles(events, power_peaks, powers, dat_det, dat_orig, time,
                  s_freq):
    """Create dict for each spindle, based on events of time points.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples, and peak frequency
    power_peaks : ndarray (dtype='float')
        peak in power spectrum for each event
    powers : ndarray (dtype='float')
        average power in power spectrum for each event
    dat_det : ndarray (dtype='float')
        vector with the data after detection-transformation (to compute peak)
    dat_orig : ndarray (dtype='float')
        vector with the raw data on which detection was performed
    time : ndarray (dtype='float')
        vector with time points
    s_freq : float
        sampling frequency

    Returns
    -------
    list of dict
        list of all the spindles, with information about start_time, peak_time,
        end_time (s), peak_val (signal units), area_under_curve
        (signal units * s), peak_freq (Hz)
    """
    i, events = _remove_duplicate(events, dat_det)
    power_peaks = power_peaks[i]

    spindles = []
    for i, one_peak, one_pwr in zip(events, power_peaks, powers):
        one_spindle = {'start': time[i[0]],
                       'end': time[i[2] - 1],
                       'peak_time': time[i[1]],
                       'peak_val_det': dat_det[i[1]],
                       'peak_val_orig': dat_orig[i[1]],
                       'dur': (i[2] - i[0]) / s_freq,
                       'auc_det': sum(dat_det[i[0]:i[2]]) / s_freq,
                       'auc_orig': sum(dat_orig[i[0]:i[2]]) / s_freq,
                       'rms_det': sqrt(mean(square(dat_det[i[0]:i[2]]))),
                       'rms_orig': sqrt(mean(square(dat_orig[i[0]:i[2]]))),
                       'power_orig': one_pwr,
                       'peak_freq': one_peak,
                       'ptp_det': ptp(dat_det[i[0]:i[2]]),
                       'ptp_orig': ptp(dat_orig[i[0]:i[2]])
                       }
        spindles.append(one_spindle)

    return spindles


def _remove_duplicate(old_events, dat):
    """Remove duplicates from the events.

    Parameters
    ----------
    old_events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    dat : ndarray (dtype='float')
        vector with the data after detection-transformation (to compute peak)

    Returns
    -------
    ndarray (dtype='int')
        vector of indices of the events to keep
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples

    Notes
    -----
    old_events is assumed to be sorted. It only checks for the start time and
    end time. When two (or more) events have the same start time and the same
    end time, then it takes the largest peak.

    There is no tolerance, indices need to be identical.
    """
    diff_events = diff(old_events, axis=0)
    dupl = where((diff_events[:, 0] == 0) & (diff_events[:, 2] == 0))[0]
    dupl += 1  # more convenient, it copies old_event first and then compares

    n_nondupl_events = old_events.shape[0] - len(dupl)
    new_events = zeros((n_nondupl_events, old_events.shape[1]), dtype='int')
    if len(dupl):
        lg.debug('Removing ' + str(len(dupl)) + ' duplicate events')

    i = 0
    indices = []
    for i_old, one_old_event in enumerate(old_events):
        if i_old not in dupl:
            new_events[i, :] = one_old_event
            i += 1
            indices.append(i_old)
        else:
            peak_0 = new_events[i - 1, 1]
            peak_1 = one_old_event[1]
            if dat[peak_0] >= dat[peak_1]:
                new_events[i - 1, 1] = peak_0
            else:
                new_events[i - 1, 1] = peak_1

    return indices, new_events


def _detect_start_end(true_values):
    """From ndarray of bool values, return intervals of True values.

    Parameters
    ----------
    true_values : ndarray (dtype='bool')
        array with bool values

    Returns
    -------
    ndarray (dtype='int')
        N x 2 matrix with starting and ending times.
    """
    neg = zeros((1), dtype='bool')
    int_values = asarray(concatenate((neg, true_values[:-1], neg)),
                         dtype='int')
    # must discard last value to avoid axis out of bounds
    cross_threshold = diff(int_values)

    event_starts = where(cross_threshold == 1)[0]
    event_ends = where(cross_threshold == -1)[0]

    if len(event_starts):
        events = vstack((event_starts, event_ends)).T

    else:
        events = None

    return events


def _select_period(detected, true_values):
    """For the detected values, we check when it goes above/below the
    selection.

    Parameters
    ----------
    detected : ndarray (dtype='int')
        N x 3 matrix with starting and ending times.
    true_values : ndarray (dtype='bool')
        array with bool values

    Returns
    -------
    ndarray (dtype='int')
        N x 2 matrix with starting and ending times, but these periods are
        usually larger than those of the input, because the selection window is
        usually more lenient (lower threshold) than the detection window.

    Notes
    -----
    Both start and end time points are inclusive (not python convention, but
    matlab convention) because these values are converted to time points later.
    """
    true_values = invert(true_values)

    for one_spindle in detected:
        # get the first time point when it goes above/below selection thres
        start_sel = where(true_values[:one_spindle[0]])[0]
        if start_sel.any():
            one_spindle[0] = start_sel[-1]

        # get the last time point when it stays above/below selection thres
        end_sel = where(true_values[one_spindle[2]:])[0] - 1
        if end_sel.any():
            one_spindle[2] += end_sel[0]

    return detected


def _merge_close(dat, events, time, min_interval):
    """Merge together events separated by less than a minimum interval.

    Parameters
    ----------
    dat : ndarray (dtype='float')
        vector with the data after selection-transformation
    events : ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    time : ndarray (dtype='float')
        vector with time points
    min_interval : float
        minimum delay between consecutive events, in seconds

    Returns
    -------
    ndarray (dtype='int')
        N x 3 matrix with start, peak, end samples
    """
    if not events.any():
        return events

    no_merge = time[events[1:, 0] - 1] - time[events[:-1, 2]] >= min_interval

    if no_merge.any():
        begs = concatenate([[events[0, 0]], events[1:, 0][no_merge]])
        ends = concatenate([events[:-1, 2][no_merge], [events[-1, 2]]])

        new_events = vstack((begs, ends)).T
    else:
        new_events = asarray([[events[0, 0], events[-1, 2]]])

    # add the location of the peak in the middle
    new_events = insert(new_events, 1, 0, axis=1)
    for i in new_events:
        if i[2] - i[0] >= 1:
            i[1] = i[0] + argmax(dat[i[0]:i[2]])

    return new_events


def _wmorlet(f0, sd, sampling_rate, ns=5):
    """Adapted from nitime

    Returns a complex morlet wavelet in the time domain

    Parameters
    ----------
        f0 : center frequency
        sd : standard deviation of frequency
        sampling_rate : samplingrate
        ns : window length in number of standard deviations

    Returns
    -------
    ndarray
        complex morlet wavelet in the time domain
    """
    st = 1. / (2. * pi * sd)
    w_sz = float(int(ns * st * sampling_rate))  # half time window size
    t = arange(-w_sz, w_sz + 1, dtype=float) / sampling_rate
    w = (exp(-t ** 2 / (2. * st ** 2)) * exp(2j * pi * f0 * t) /
         sqrt(sqrt(pi) * st * sampling_rate))
    return w


def _realwavelets(s_freq, freqs, dur, width):
    """Create real wavelets, for UCSD.

    Parameters
    ----------
    s_freq : int
        sampling frequency
    freqs : ndarray
        vector with frequencies of interest
    dur : float
        duration of the wavelets in s
    width : float
        parameter controlling gaussian shape

    Returns
    -------
    ndarray
        wavelets
    """
    x = arange(-dur / 2, dur / 2, 1 / s_freq)
    wavelets = empty((len(freqs), len(x)))

    g = exp(-(pi * x ** 2) / width ** 2)

    for i, one_freq in enumerate(freqs):
        y = cos(2 * pi * x * one_freq)
        wavelets[i, :] = y * g

    return wavelets



"""Module to detect arousals
"""

from logging import getLogger
from numpy import abs, argmin, asarray, hstack, mean, sum, vstack, where, zeros
from scipy.signal import spectrogram

try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QProgressDialog
except ImportError:
    pass

from .spindle import within_duration, remove_straddlers
from ..graphoelement import Arousals

lg = getLogger(__name__)


class DetectArousal:
    """Design slow wave detection on a single channel.

    Parameters
    ----------
    method : str
        one of the predefined methods
    freq_band : tuple of (float or None)
        frequency band of interest in Hz
    spectrogram : dict
        'dur': float
            window length in sec 
        'overlap': float
            ratio of overlap between consecutive windows 
        'detrend': str
            'constant', 'linear' or False
    det_thresh : float
        minimum factor increase of mean frequency between consecutive windows
    min_interval : float
        minimum duration between consecutive arousals, in sec
    duration : tuple of float
        min and max duration of arousals
    """
    def __init__(self, method='HouseDetector', duration=None):

        self.method = method

        if method == 'HouseDetector':
            self.freq_band1 = (5, None)
            self.freq_band2 = (0.2, None)
            self.spectrogram = {'dur': 1,
                                'overlap': 0.5,
                                'detrend': 'linear'}
            self.det_thresh = 1.2
            self.det_thresh_end = 1.1
            self.min_interval = 10
            self.duration = (3, 30)

        else:
            raise ValueError('Unknown method')
            
        if duration is None:
            self.duration = (self.min_dur, self.max_dur)       
        else:
            self.duration = duration

    def __repr__(self):
        return ('detsw_{0}_{1:04.2f}-{2:04.2f}Hz'
                ''.format(self.method, *self.det_filt['freq']))

    def __call__(self, data, parent=None):
        """Detect slow waves on the data.

        Parameters
        ----------
        data : instance of Data
            data used for detection
        parent : QWidget
            for use with GUI, as parent widget for the progress bar
        
        Returns
        -------
        instance of graphoelement.Arousals
            description of the detected arousals
        """
        if parent is not None:
            progress = QProgressDialog('Finding arousals', 'Abort', 
                                       0, data.number_of('chan')[0], parent)
            progress.setWindowModality(Qt.ApplicationModal)
            
        arousal = Arousals()
        arousal.chan_name = data.axis['chan'][0]

        all_arousals = []
        for i, chan in enumerate(data.axis['chan'][0]):
            
            lg.info('Detecting arousals on chan %s', chan)
            time = hstack(data.axis['time'])
            dat_orig = hstack(data(chan=chan))

            if 'HouseDetector' in self.method:
                arou_in_chan = detect_HouseDetector(dat_orig, data.s_freq, time,
                                                  self)

            else:
                raise ValueError('Unknown method')

            for ar in arou_in_chan:
                ar.update({'chan': chan})
            all_arousals.extend(arou_in_chan)
            
            if parent is not None:
                progress.setValue(i)
                if progress.wasCanceled():
                    return
            # end of loop over chan

        arousal.events = sorted(all_arousals, key=lambda x: x['start'])

        if parent is not None:
            progress.setValue(i + 1)

        return arousal

def detect_HouseDetector(dat_orig, s_freq, time, opts):
    """House arousal detection.

    Parameters
    ----------
    dat_orig : ndarray (dtype='float')
        vector with the data for one channel
    s_freq : float
        sampling frequency
    time : ndarray (dtype='float')
        vector with the time points for each sample
    opts : instance of 'DetectSlowWave'
        'duration' : tuple of float
            min and max duration of arousal

    Returns
    -------
    list of dict
        list of detected arousals
    float
        arousal density, per 30-s epoch
    """  
    nperseg = int(opts.spectrogram['dur'] * s_freq)
    overlap = opts.spectrogram['overlap']
    noverlap = int(overlap * nperseg)
    detrend = opts.spectrogram['detrend']
    min_interval = int(opts.min_interval * s_freq)
    
    sf, t, dat_det = spectrogram(dat_orig, 
                                 fs=s_freq, 
                                 nperseg=nperseg, 
                                 noverlap=noverlap, 
                                 detrend=detrend)
    freq1 = opts.freq_band1
    freq2 = opts.freq_band2
    f0 = asarray([abs(freq1[0] - x) for x in sf]).argmin() if freq1[0] else None
    f1 = asarray([abs(freq1[1] - x) for x in sf]).argmin() if freq1[1] else None
    f2 = asarray([abs(freq2[1] - x) for x in sf]).argmin() if freq2[1] else None
    f3 = asarray([abs(freq2[1] - x) for x in sf]).argmin() if freq2[1] else None
    
    dat_eq1 = zeros(dat_det.shape[1])
    dat_eq2 = zeros(dat_det.shape[1])
    for i in range(dat_det.shape[1]):
        dat_eq1[i] = splitpoint(dat_det[f0:f1, i], sf[f0:f1])
        dat_eq2[i] = splitpoint(dat_det[f2:f3, i], sf[f2:f3])
        
    dat_acc = dat_eq1[1:] / dat_eq1[:-1]
    starts = dat_acc >= opts.det_thresh
    print(f'starts: {sum(starts)}')
    print(f'1.01: {sum(dat_acc >= 1.01)}')
    print(f'1.02: {sum(dat_acc >= 1.02)}')
    print(f'1.05: {sum(dat_acc >= 1.05)}')
    print(f'1.1: {sum(dat_acc >= 1.1)}')
    print(f'1.2: {sum(dat_acc >= 1.2)}')
    print(f'1.3: {sum(dat_acc >= 1.3)}')
    print(f'1.4: {sum(dat_acc >= 1.4)}')
    print(f'1.5: {sum(dat_acc >= 1.5)}')
    print(f'1.75: {sum(dat_acc >= 1.75)}')
    print(f'2: {sum(dat_acc >= 2)}')
    print(f'2.5: {sum(dat_acc >= 2.5)}')
    print(f'3: {sum(dat_acc >= 3)}')
    print(f'5: {sum(dat_acc >= 5)}')
    print(f'10: {sum(dat_acc >= 10)}')
    
    if starts.any():
        new_starts = asarray(zeros(len(starts)), dtype=bool)
        ends = asarray(zeros(len(starts) - 1), dtype=bool)
        iter_len = len(starts) - 2
        i = 0
        while i <= iter_len:
            if starts[i]:
                for j, k in enumerate(dat_eq2[i + 2:-1]):
                    if k < dat_eq2[i] * opts.det_thresh_end:
                        new_starts[i] = True
                        ends[i + j + 1] = True
                        break
                i += j + min_interval
            else:
                i += 1
        
        if sum(new_starts) > sum(ends): # a start without an end
            ends[-1] = True
        
        events = vstack((where(new_starts == True)[0] + 1,
                         where(ends == True)[0] + 2)).T
        if overlap: 
            events = events - int(1 / 2 / overlap) # from win centre to win start
        events = events * (nperseg - noverlap) # upsample
        print(f'n_events before dur = {events.shape}')
        events = within_duration(events, time, opts.duration)
        print(f'n_events after dur = {events.shape}')
        events = remove_straddlers(events, time, s_freq)
        print(f'n_events after strad = {events.shape}')
    
        ar_in_chan = make_arousals(events, time, s_freq)
        
    else:
        lg.info('No arousals found')
        ar_in_chan = []

    return ar_in_chan


def splitpoint(a, sf):
    c1 = a.cumsum()
    c2 = a[::-1].cumsum()[::-1]
    split = argmin(abs(c1-c2))
    return sf[split]

def make_arousals(events, time, s_freq):
    """Create dict for each arousal, based on events of time points.

    Parameters
    ----------
    events : ndarray (dtype='int')
        N x 5 matrix with start, end samples
    data : ndarray (dtype='float')
        vector with the data
    time : ndarray (dtype='float')
        vector with time points
    s_freq : float
        sampling frequency

    Returns
    -------
    list of dict
        list of all the arousals, with information about start, end, 
        duration (s),
    """
    arousals = []
    for ev in events:
        one_ar = {'start': time[ev[0]],
                  'end': time[ev[1] - 1],
                  'dur': (ev[1] - ev[0]) / s_freq,
                  }
        arousals.append(one_ar)

    return arousals



from numpy import isnan

from wonambi import Dataset

from .paths import mff_file


d = Dataset(mff_file)

def test_mff_read():
    d.read_data(begtime=10, endtime=20)


def test_mff_before_start():
    data = d.read_data(begsam=-100, endsam=10)
    assert isnan(data.data[0][0, 0])


def test_mff_after_end():
    n_samples = d.header['n_samples']
    data = d.read_data(begsam=n_samples - 100, endsam=n_samples + 100)
    assert isnan(data.data[0][0, -1])


def test_mff_markers():
    markers = d.read_markers()
    assert len(markers) == 8



"""Module contains all the exceptions

"""

class UnrecognizedFormat(Exception):
    """Could not recognize the format of the file for channels.

    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class MissingDependency():
    """Class to handle missing dependencies in an elegant and consistent way.

    Examples
    --------
    When importing a function from an optional dependency, use this construct:

    >>> try:
    >>>     from optdep import function
    >>> except ImportError as err:
    >>>     function = MissingDependency(err)

    and the use ``function`` in your code normally. It'll raise an ImportError
    only if the code calls that specific function.
    """
    def __init__(self, error):
        self.error = error

    def __call__(self, *args, **kwargs):
        raise ImportError(f"You need to install the optional dependency '{self.error.name}' to run this function") from self.error



from numpy import (append,
                   arange,
                   array,
                   cumsum,
                   empty,
                   isnan,
                   nan
                   )
from numpy.testing import assert_array_equal
from pytest import raises

from wonambi.ioeeg.utils import _select_blocks


BLOCKS = array([5, 11, 6, 7, 12])


def test_select_block_first_block():
    _assert_begsam_endsam(0, 2)


def test_select_block_second_block():
    _assert_begsam_endsam(5, 9)


def test_select_block_multiple_blocks():
    _assert_begsam_endsam(3, 38)


def test_select_block_empty():
    dat = _assert_begsam_endsam(1, 1)
    assert dat.shape[0] == 0


def test_select_block_begsam_before():
    dat = _assert_begsam_endsam(-10, 1)
    assert dat.shape[0] == 11
    assert isnan(dat[0])


def test_select_block_endsam_after():
    dat = _assert_begsam_endsam(10, 45)
    assert dat.shape[0] == 35
    assert isnan(dat[-1])


def test_select_block_endsam_before():
    with raises(StopIteration):
        next(_select_blocks(BLOCKS, -10, -5))


def test_select_block_begsam_after():
    with raises(StopIteration):
        next(_select_blocks(BLOCKS, 50, 55))


def _assert_begsam_endsam(begsam, endsam):

    dat_on_disk, intervals = _generate_dat_on_disk()

    expected_dat = arange(begsam, endsam).astype('float')
    expected_dat[(expected_dat < 0) & (expected_dat > intervals[-1])] = nan

    dat = empty(endsam - begsam)
    dat.fill(nan)

    for i_dat, blk, i_blk in _select_blocks(BLOCKS, begsam, endsam):
        # print('{: 3d}-{: 3d} = {: 3d}: {: 3d}-{: 3d}'.format(i_dat[0], i_dat[1], blk, i_blk[0], i_blk[1]))
        dat[i_dat[0]:i_dat[1]] = dat_on_disk[blk][i_blk[0]:i_blk[1]]

    return dat

    assert_array_equal(expected_dat, dat)


def _generate_dat_on_disk():
    intervals = cumsum(append(0, BLOCKS))

    dat_on_disk = []
    for i in range(len(intervals) - 1):
        dat_on_disk.append(arange(intervals[i], intervals[i + 1]))

    return dat_on_disk, intervals



"""Class to import HRV RRI data in text format.
"""
from logging import getLogger
from numpy import arange, cumsum, empty, float64, interp, newaxis
from scipy.interpolate import splev, splrep
from datetime import datetime, timedelta

from .utils import DEFAULT_DATETIME

lg = getLogger(__name__)


class LyonRRI:
    """Class to read text format RR interval data from HRVanalysis export 
    (INSERM/Lyon). Returns one channel only.

    Parameters
    ----------
    rec_dir : path to record directory
        the folder containing the record
        
    Notes
    -----
    With code adapted from Rhenan Bartels: https://github.com/rhenanbartels/hrv
    """
    def __init__(self, rec_dir):
        lg.info('Reading ' + str(rec_dir))
        self.filename = rec_dir
        self.s_freq = None
        self.hdr = self.return_hdr()
        
        self.dig_min = 0
        self.dig_max = 3000
        self.phys_min = 0
        self.phys_max = 3000
        
        self.rri = self.return_rri(0, self.hdr[4])
        self.time = self.create_time(self.rri)
        self.time_interp = None
        self.rri_interp = None
        
    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            the full header
        """
        hdr = {}            
        hdr['s_freq'] = self.s_freq
        hdr['chan_name'] = ['RRi']
        
        with open(self.filename, 'rt') as f:
            head = [next(f) for x in range(12)]
            hdr['subj_id'] = head[0][11:-3]
            hdr['start_time'] = DEFAULT_DATETIME
            hdr['recorder'] = head[2][10:]
            hdr['s_freq_ecg'] = int(head[3][4:]) # ECG sampling frequency
            t = datetime.strptime(head[4][16:24], '%H:%M:%S')
            hdr['total_dur'] = timedelta(hours=t.hour, minutes=t.minute, 
               seconds=t.second)
            hdr['export_date'] = DEFAULT_DATETIME
            hdr['data_type'] = head[10][11:]
            
            for i, _ in enumerate(f):
                pass
            hdr['n_samples'] = i
        
        output = (hdr['subj_id'], hdr['start_time'], hdr['s_freq'], 
                  hdr['chan_name'], hdr['n_samples'], hdr)
        
        return output
    
    def return_dat(self, chan, begsam, endsam):
        if self.rri_interp is None:
            raise ValueError('RRi has not been interpolated.')
                        
        return self.rri_interp[newaxis, begsam:endsam]

    def return_markers(self):
        """There are no markers in this format.
        """
        return []
    
    def create_time(self, rri):
        time = (cumsum(rri) / 1000.0)        
        return time - time[0]
    
    def return_rri(self, begsam, endsam):
        """Return raw, irregularly-timed RRI."""
        interval = endsam - begsam
        dat = empty(interval)
        k = 0
            
        with open(self.filename, 'rt') as f:
            [next(f) for x in range(12)]
            
            for j, datum in enumerate(f):
                
                if begsam <= j < endsam:
                    dat[k] = float64(datum[:datum.index('\t')])
                    k += 1
                    if k == interval:
                        break
                    
        return dat
    
    def interpolate(self, s_freq=4, interp_method='cubic'):
        rri = self.rri
        irregular_time = self.time
        step = 1 / float(s_freq)
        regular_time = arange(0, irregular_time[-1] + step, step)
        
        if interp_method == 'cubic':
            tck = splrep(irregular_time, rri, s=0)
            rri_interp = splev(regular_time, tck, der=0)
            
        elif interp_method == 'linear':
            rri_interp = interp(regular_time, irregular_time, rri)
            
        self.time_interp = regular_time
        self.rri_interp = rri_interp
        self.s_freq = s_freq



"""Module has information about the datasets, not data.

"""
from datetime import timedelta, datetime
from math import ceil
from logging import getLogger
from pathlib import Path

from numpy import arange, asarray, concatenate, empty, int64, zeros, ndarray

from .ioeeg import (Abf,
                    Edf,
                    Ktlx,
                    BlackRock,
                    EgiMff,
                    EEGLAB,
                    FieldTrip,
                    BrainVision,
                    Moberg,
                    Wonambi,
                    Micromed,
                    BCI2000,
                    OpenEphys,
                    Text,
                    BIDS,
                    LyonRRI,
                    )
from .ioeeg.bci2000 import _read_header_length
from .datatype import ChanTime
from .utils import UnrecognizedFormat


lg = getLogger('wonambi')


def _convert_time_to_sample(abs_time, dataset):
    """Convert absolute time into samples.

    Parameters
    ----------
    abs_time : dat
        if it's int or float, it's assumed it's s;
        if it's timedelta, it's assumed from the start of the recording;
        if it's datetime, it's assumed it's absolute time.
    dataset : instance of wonambi.Dataset
        dataset to get sampling frequency and start time

    Returns
    -------
    int
        sample (from the starting of the recording).
    """
    if isinstance(abs_time, datetime):
        abs_time = abs_time - dataset.header['start_time']

    if not isinstance(abs_time, timedelta):
        try:
            abs_time = timedelta(seconds=float(abs_time))
        except TypeError as err:
            if isinstance(abs_time, int64):
                # timedelta and int64: http://bugs.python.org/issue5476
                abs_time = timedelta(seconds=int(abs_time))
            else:
                raise err

    sample = int(ceil(abs_time.total_seconds() * dataset.header['s_freq']))
    return sample


def detect_format(filename):
    """Detect file format.

    Parameters
    ----------
    filename : str or Path
        name of the filename or directory.

    Returns
    -------
    class used to read the data.

    list : indices of sessions
    """
    sessions = [1, ]  # start counting from 1
    filename = Path(filename)

    if filename.is_dir():
        if list(filename.glob('*.stc')) and list(filename.glob('*.erd')):
            return Ktlx, sessions
        elif (filename / 'patient.info').exists():
            return Moberg, sessions
        elif (filename / 'info.xml').exists():
            return EgiMff, sessions
        elif list(filename.glob('*.openephys')):
            sessions = _count_openephys_sessions(filename)
            return OpenEphys, sessions
        elif list(filename.glob('*.txt')):
            return Text, sessions
        else:
            raise UnrecognizedFormat('Unrecognized format for directory ' +
                                     str(filename))
    else:
        if filename.suffix == '.won':
            return Wonambi, sessions

        if filename.suffix.lower() == '.trc':
            return Micromed, sessions

        if filename.suffix == '.set':
            return EEGLAB, sessions

        if filename.suffix in ['.edf', '.rec']:
            return Edf, sessions

        if filename.suffix == '.abf':
            return Abf, sessions

        if filename.suffix == '.vhdr' or filename.suffix == '.eeg':
            return BrainVision, sessions

        if filename.suffix == '.dat':  # very general
            try:
                _read_header_length(filename)

            except (AttributeError, ValueError):  # there is no HeaderLen
                pass

            else:
                return BCI2000, sessions

        with filename.open('rb') as f:
            file_header = f.read(8)
            if file_header in (b'NEURALCD', b'NEURALSG', b'NEURALEV'):
                return BlackRock, sessions
            elif file_header[:6] == b'MATLAB':  # we might need to read more
                return FieldTrip, sessions

        if filename.suffix.lower() == '.txt':
            with filename.open('rt') as f:
                first_line = f.readline()
                if '.rr' in first_line[-4:]:
                    return LyonRRI, sessions

        else:
            raise UnrecognizedFormat('Unrecognized format for file ' +
                                     str(filename))


class Dataset:
    """Contain specific information and methods, associated with a dataset.

    Parameters
    ----------
    filename : str or Path
        name of the file
    IOClass : class
        one of the classes of wonambi.ioeeg
    bids : bool
        whether you give precedence to the information stored in the accompanying
        files which are in the BIDS format

    Attributes
    ----------
    filename : str
        name of the file
    IOClass : class
        format of the file
    header : dict
        - subj_id : str
            subject identification code
        - start_time : datetime
            start time of the dataset. If it cannot get the start time from the
            header, wonambi reads it from the timestamp of the file
        - s_freq : float
            sampling frequency
        - chan_name : list of str
            list of all the channels
        - n_samples : int
            number of samples in the dataset
        - orig : dict
            additional information taken directly from the header
    dataset : instance of a class which depends on format,
        this requires at least three attributes:
          - filename
          - return_hdr
          - return_dat

    Notes
    -----
    There is a difference between Dataset.filename and Dataset.dataset.filename
    because the former is where the file that you want to read (the argument),
    while the latter is the file that you really read. There might be
    differences, for example, if the argument points to a file within a
    directory, or if the file is mapped to memory.
    """
    def __init__(self, filename, IOClass=None, session=None, bids=False):
        self.filename = Path(filename)

        if bids:
            IOClass = BIDS

        if IOClass is not None:
            self.IOClass = IOClass
        else:
            self.IOClass, sessions = detect_format(filename)

        if self.IOClass in (OpenEphys, ):
            if session is None:
                session = 1
                if len(sessions) > 1:
                    lg.warning(f'Multiple sessions in the dataset, selecting the first one. You can specify the session with "session="')

            lg.debug(f'Reading session {session}')
            self.dataset = self.IOClass(self.filename, session=session)

        else:
            self.dataset = self.IOClass(self.filename)

        output = self.dataset.return_hdr()
        hdr = {}
        hdr['subj_id'] = output[0]
        hdr['start_time'] = output[1]
        if hdr['start_time'] is None:
            hdr['start_time'] = datetime.fromtimestamp(self.filename.stat().st_mtime)
        hdr['s_freq'] = output[2]
        hdr['chan_name'] = output[3]
        hdr['n_samples'] = output[4]
        hdr['orig'] = output[5]
        self.header = hdr

    def read_markers(self, **kwargs):
        """Return the markers. You can add optional arguments that will be
        passed to the method specific for each datafile.
        """
        return self.dataset.return_markers(**kwargs)

    def read_videos(self, begtime=None, endtime=None):
        """Return list of videos with start and end times for a period.

        Parameters
        ----------
        begtime : int or timedelta or datetime or list
            start of the data to read;
            if it's int, it's assumed it's s;
            if it's timedelta, it's assumed from the start of the recording;
            if it's datetime, it's assumed it's absolute time.
            It can also be a list of any of the above type.
        endtime : int or timedelta or datetime
            end of the data to read;
            if it's int, it's assumed it's s;
            if it's timedelta, it's assumed from the start of the recording;
            if it's datetime, it's assumed it's absolute time.
            It can also be a list of any of the above type.

        Returns
        -------
        list of path
            list of absolute paths (as str) to the movie files
        float
            time in s from the beginning of the first movie when the part of
            interest starts
        float
            time in s from the beginning of the last movie when the part of
            interest ends

        Raises
        ------
        OSError
            when there are no video files at all
        IndexError
            when there are video files, but the interval of interest is not in
            the list of files.
        """
        if isinstance(begtime, datetime):
            begtime = begtime - self.header['start_time']
        if isinstance(begtime, timedelta):
            begtime = begtime.total_seconds()
        if isinstance(endtime, datetime):
            endtime = endtime - self.header['start_time']
        if isinstance(endtime, timedelta):
            endtime = endtime.total_seconds()

        videos = self.dataset.return_videos(begtime, endtime)
        """
        try
        except AttributeError:
            lg.debug('This format does not have video')
            videos = None
        """
        return videos

    def read_data(self, chan=None, begtime=None, endtime=None, begsam=None,
                  endsam=None, events=None, pre=1, post=1, s_freq=None):
        """Read the data and creates a ChanTime instance

        Parameters
        ----------
        chan : list of strings
            names of the channels to read
        begtime : int or timedelta or datetime or list
            start of the data to read;
            if it's int or float, it's assumed it's s;
            if it's timedelta, it's assumed from the start of the recording;
            if it's datetime, it's assumed it's absolute time.
            It can also be a list of any of the above type.
        endtime : int or timedelta or datetime
            end of the data to read;
            if it's int or float, it's assumed it's s;
            if it's timedelta, it's assumed from the start of the recording;
            if it's datetime, it's assumed it's absolute time.
            It can also be a list of any of the above type.
        begsam : int
            first sample (this sample will be included)
        endsam : int
            last sample (this sample will NOT be included)
        events : list of int or of timedelta or of datetime
            list of the onset time of the events of interest.
            This option is useful if you want to run a trial-based analysis.
        pre : float
            only when "events" is specified, the amount of data before each
            event to be included (in s). Use a positive number to indicate
            the time before the event.
        post : float
            only when "events" is specified, the amount of data after each
            event to be included (in s).
        s_freq : int
            sampling frequency of the data

        Returns
        -------
        An instance of ChanTime

        Notes
        -----
        begsam and endsam follow Python convention, which starts at zero,
        includes begsam but DOES NOT include endsam.

        If begtime and endtime are a list, the two lists should have the same
        length and the data will be stored in trials.

        If neither begtime or begsam are specified, it starts from the first
        sample. If neither endtime or endsam are specified, it reads until the
        end.

        The time axis will indicate the time in seconds from data.start_time,
        unless you specify "events". In that case, time will run from -"pre" to
        +"post".
        """
        data = ChanTime()
        data.start_time = self.header['start_time']
        data.s_freq = s_freq = s_freq if s_freq else self.header['s_freq']

        if chan is None:
            chan = self.header['chan_name']
        if not (isinstance(chan, list) or isinstance(chan, tuple) or isinstance(chan, ndarray)):
            raise TypeError('Parameter "chan" should be a list')
        else:
            chan = list(chan)  # make sure it's a list
        add_ref = False
        if '_REF' in chan:
            add_ref = True
            chan[:] = [x for x in chan if x != '_REF']
        idx_chan = [self.header['chan_name'].index(x) for x in chan]

        if begtime is None and begsam is None:
            begsam = 0
        if endtime is None and endsam is None:
            endsam = self.header['n_samples']

        if events is not None:
            eventssam = self._convert_to_list_with_samples(events)
            presam = int(pre * s_freq)
            postsam = int(post * s_freq)
            begsam = [event - presam for event in eventssam]
            endsam = [event + postsam for event in eventssam]

            event_t = arange(-presam, postsam) / s_freq

        begsam = self._convert_to_list_with_samples(begtime, begsam)
        endsam = self._convert_to_list_with_samples(endtime, endsam)

        if len(begsam) != len(endsam):
            raise ValueError('There should be the same number of start and ' +
                             'end point')
        n_trl = len(begsam)

        data.axis['chan'] = empty(n_trl, dtype='O')
        data.axis['time'] = empty(n_trl, dtype='O')
        data.data = empty(n_trl, dtype='O')

        for i, one_begsam, one_endsam in zip(range(n_trl), begsam, endsam):
            dataset = self.dataset
            #lg.debug('begsam {0: 6}, endsam {1: 6}'.format(one_begsam,
            #         one_endsam))
            dat = dataset.return_dat(idx_chan, one_begsam, one_endsam)
            chan_in_dat = chan

            if add_ref:
                zero_ref = zeros((1, one_endsam - one_begsam))
                dat = concatenate((dat, zero_ref), axis=0)
                chan_in_dat.append('_REF')

            data.data[i] = dat
            data.axis['chan'][i] = asarray(chan_in_dat, dtype='U')
            if events is not None:
                data.axis['time'][i] = event_t
            else:
                data.axis['time'][i] = arange(one_begsam, one_endsam) / s_freq

        return data

    def _convert_to_list_with_samples(self, times=None, samples=None):
        """Convenience function to convert the input into a list of samples"""
        if times is not None:
            if not isinstance(times, list):
                times = [times]
            samples = []
            for one_time in times:
                samples.append(_convert_time_to_sample(one_time, self))

        if not isinstance(samples, list):
            samples = [samples]

        return samples


def _count_openephys_sessions(filename):
    """Open-ephys can have multiple sessions. We count how many files are in
    the format:
      - Continuous_Data.openephys
      - Continuous_Data_2.openephys
      - Continuous_Data_3.openephys
    """
    sessions = []
    for f in filename.glob('*.openephys'):
        session_number = f.stem[16:]
        if session_number == '':
            sessions.append(1)
        else:
            sessions.append(int(session_number))

    return sorted(sessions)
